/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.displaysimulator;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;

import org.disperanto.displaysimulator.data.Display;
import org.disperanto.displaysimulator.data.DisplayController;
import org.disperanto.displaysimulator.data.DisplayCreator;
import org.disperanto.displaysimulator.listeners.IAvailableDisplayListener;
import org.disperanto.displaysimulator.ui.controllers.BrightnessViewController;
import org.disperanto.displaysimulator.ui.controllers.CommunicationTimeoutViewController;
import org.disperanto.displaysimulator.ui.controllers.DisplayControllerViewController;
import org.disperanto.displaysimulator.ui.controllers.ImageViewController;
import org.disperanto.displaysimulator.ui.controllers.NotificationViewController;
import org.disperanto.displaysimulator.ui.controllers.PropertiesViewController;
import org.disperanto.displaysimulator.ui.controllers.StatusViewController;
import org.disperanto.displaysimulator.ui.models.DisplayModel;
import org.disperanto.displaysimulator.ui.models.LoggingModel;
import org.disperanto.displaysimulator.ui.models.NotificationModel;
import org.disperanto.displaysimulator.ui.models.PropertiesModel;
import org.disperanto.displaysimulator.ui.models.StatusModel;
import org.disperanto.displaysimulator.ui.panels.BrightnessPanel;
import org.disperanto.displaysimulator.ui.panels.CommunicationTimeoutPanel;
import org.disperanto.displaysimulator.ui.panels.DisplayControllerPanel;
import org.disperanto.displaysimulator.ui.panels.DisplayListPanel;
import org.disperanto.displaysimulator.ui.panels.ImagePanel;
import org.disperanto.displaysimulator.ui.panels.LoggingPanel;
import org.disperanto.displaysimulator.ui.panels.NotificationPanel;
import org.disperanto.displaysimulator.ui.panels.PropertiesPanel;
import org.disperanto.displaysimulator.ui.panels.StatusPanel;
import org.disperanto.displaysimulator.ui.panels.TextPanel;

import net.miginfocom.swing.MigLayout;

/**
 * Main dialog for Display simulator for the Disperanto.
 */
public class SimulatorMainDialog extends JFrame {
    private static final String VERSION = "v1.2";

    private DisplayController controller;
    private LoggingModel loggingModel;

    private JTabbedPane tabbedPane;
    private ImagePanel imagePanel;
    private TextPanel textPanel;

    private ImageViewController imageController;
    private StatusViewController statusController;
    private PropertiesViewController propertiesController;
    private NotificationViewController notificationController;
    private CommunicationTimeoutViewController timeoutController;
    private BrightnessViewController brightnessController;
    
    private PropertiesModel displayPropertiesModel;

    /**
     * Constructor
     * 
     * @param title Application title
     * @param controller controller for the displays
     * @param loggingModel Model for logging panel
     */
    public SimulatorMainDialog(String title, DisplayController controller, LoggingModel loggingModel) {
        super(title);
        loggingModel.log(title + " " + VERSION);

        this.controller = controller;
        this.loggingModel = loggingModel;

        addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent we) {
                System.exit(0);
                super.windowClosing(we);
            }
        });

        final JPanel panel = new JPanel(new MigLayout());
        panel.add(createMainPanel(), "center");
        setContentPane(panel);

        controller.addAvailableDisplayListener(new AvailableDisplayListener());

        // Add initial displays
        DisplayCreator displayCreator = new DisplayCreator();
        controller.addDisplay(displayCreator.createDisplay(6));
        controller.addDisplay(displayCreator.createDisplay(21));
        controller.addDisplay(displayCreator.createDisplay(22));

        pack();
        setSize(new Dimension(1280, 768));
        setLocationRelativeTo(null);
    }

    private Component createMainPanel() {
        // create panel for whole window
        final JPanel panel = new JPanel(new MigLayout("", "[grow][][][]", "[grow][]"));

        panel.add(createLoggingPanel(), "south");
        panel.add(createDisplayPanel(), "cell 0 0, top, growx, wrap");
        panel.add(createBrightnessPanel(), "cell 0 1, grow");
        panel.add(createDisplayPropertiesPanel(), "cell 1 0 1 2, growy");
        panel.add(createStatusPanel(), "cell 2 0, growy, wrap");
        panel.add(createTimeoutPanel(), "cell 2 1, growx");
        panel.add(createNotificationPanel(), "cell 3 0 1 2 , growy");

        JPanel topPanel = new JPanel(new MigLayout("insets 0", "[grow]", "[][grow]"));
        topPanel.add(createDisplayControllerPanel(), "grow, wrap");
        topPanel.add(createDisplayList(), "grow");
        panel.add(topPanel, "north");

        displayPropertiesModel.addListener(imagePanel.getFontPreviewPanel());
        displayPropertiesModel.addListener(imageController);
        displayPropertiesModel.addListener(statusController);

        return panel;
    }

    private JPanel createDisplayControllerPanel() {
        DisplayControllerPanel controllerPanel = new DisplayControllerPanel();

        controllerPanel.addListener(new DisplayControllerViewController(controllerPanel, controller));

        return controllerPanel;
    }

    private JPanel createDisplayList() {
        DisplayListPanel displayList = new DisplayListPanel(controller);
        displayList.addSelectionListener(statusController);
        displayList.addSelectionListener(propertiesController);
        displayList.addSelectionListener(notificationController);
        displayList.addSelectionListener(imageController);
        displayList.addSelectionListener(timeoutController);
        displayList.addSelectionListener(brightnessController);
        return displayList;
    }

    private JPanel createLoggingPanel() {
        LoggingPanel loggingPanel = new LoggingPanel(loggingModel);

        final JPanel logging = new JPanel(new MigLayout("insets 10 0 0 0", "[grow]", "[]"));
        logging.add(loggingPanel.getEditor(), "grow");

        return logging;
    }

    private JPanel createDisplayPanel() {
        JPanel displayPanel = new JPanel(new MigLayout("insets 0 0 0 0", "[grow]"));
        displayPanel.setBorder(BorderFactory.createTitledBorder("Display"));

        imagePanel = new ImagePanel();
        textPanel = new TextPanel();

        DisplayModel displayModel = new DisplayModel();
        displayModel.addListener(imagePanel.getFontPreviewPanel());
        displayModel.addListener(textPanel);

        imageController = new ImageViewController(imagePanel, displayModel, controller);
        imagePanel.addActionListener(imageController);
        imagePanel.addChangeListener(imageController);

        tabbedPane = new JTabbedPane();
        tabbedPane.add("Image", imagePanel);
        tabbedPane.add("Text", textPanel);
        displayPanel.add(tabbedPane, "grow");
        return displayPanel;
    }

    private JPanel createNotificationPanel() {
        NotificationPanel notificationPanel = new NotificationPanel();

        NotificationModel notificationModel = new NotificationModel();
        notificationModel.addListener(notificationPanel);

        notificationController = new NotificationViewController(notificationPanel, notificationModel, controller);
        notificationPanel.addListener(notificationController);

        return notificationPanel;
    }

    private JPanel createDisplayPropertiesPanel() {
        PropertiesPanel displayPropertiesPanel = new PropertiesPanel();

        displayPropertiesModel = new PropertiesModel();
        displayPropertiesModel.addListener(displayPropertiesPanel);

        propertiesController = new PropertiesViewController(displayPropertiesPanel, displayPropertiesModel, controller);
        displayPropertiesPanel.addListener(propertiesController);

        return displayPropertiesPanel;
    }

    private JPanel createStatusPanel() {
        StatusPanel statusPanel = new StatusPanel();

        StatusModel statusModel = new StatusModel();
        statusModel.addListener(statusPanel);

        statusController = new StatusViewController(statusPanel, statusModel, controller);
        statusPanel.addActionListener(statusController);
        statusPanel.addItemListener(statusController);

        return statusPanel;
    }

    private JPanel createTimeoutPanel() {
        CommunicationTimeoutPanel timeoutPanel = new CommunicationTimeoutPanel();

        timeoutController = new CommunicationTimeoutViewController(timeoutPanel, controller);

        return timeoutPanel;
    }

    private JPanel createBrightnessPanel() {
        BrightnessPanel brightnessPanel = new BrightnessPanel();

        brightnessController = new BrightnessViewController(brightnessPanel, controller);

        return brightnessPanel;
    }

    private class AvailableDisplayListener implements IAvailableDisplayListener {

        @Override
        public void onDisplayAdded(Display display) {
            display.addDisplayListener(imageController);
            display.addClearNotificationListener(notificationController);
            display.addTimeoutListener(timeoutController);
            display.addBrightnessTableListener(brightnessController);
        }

        @Override
        public void onDisplayRemoved(Display display) {
            display.removeDisplayListener(imageController);
            display.removeClearNotificationListener(notificationController);
            display.removeTimeoutListener(timeoutController);
            display.removeBrightnessTableListener(brightnessController);
        }

    }

}
