/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.displaysimulator.data;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;

import org.disperanto.data.TdpBrightnessTable;
import org.disperanto.data.TdpDiagnostics;
import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.data.TdpExternalLighting;
import org.disperanto.data.TdpNotification;
import org.disperanto.data.TdpSetText;
import org.disperanto.data.TdpStatus;
import org.disperanto.data.TdpTimeoutSettings;
import org.disperanto.displaysimulator.listeners.IBrightnessTableListener;
import org.disperanto.displaysimulator.listeners.IClearNotificationsListener;
import org.disperanto.displaysimulator.listeners.ICommunicationSettingsListener;
import org.disperanto.displaysimulator.listeners.IDisplayListener;
import org.disperanto.enums.ETdpLightingState;
import org.disperanto.enums.ETdpTimeoutMode;

/**
 * Display
 */
public class Display implements AutoCloseable {

    private int address;
    private TdpDisplayProperties displayProperties;
    private TdpStatus status = new TdpStatus();
    private TdpTimeoutSettings timeoutSettings = new TdpTimeoutSettings();
    private TdpBrightnessTable brightnessTable = new TdpBrightnessTable();
    private TdpExternalLighting externalLighting = new TdpExternalLighting();
    private TdpDiagnostics diagnostics = new TdpDiagnostics();
    private List<TdpNotification> notifications = new ArrayList<TdpNotification>();

    private ImageBanks banks = new ImageBanks();
    private TdpSetText textLines = new TdpSetText();
    private ShowInfo showInfo = new ShowInfo();

    private final List<ICommunicationSettingsListener> timeoutListeners = new CopyOnWriteArrayList<>();
    private final List<IBrightnessTableListener> brightnessListeners = new CopyOnWriteArrayList<>();
    private final List<IClearNotificationsListener> clearListeners = new CopyOnWriteArrayList<>();
    private final List<IExternalLightingListener> externalLightingListeners = new CopyOnWriteArrayList<>();

    /**
     * Constructor
     *
     * @param address Address of display
     * @param properties Properties of display
     * @param status Status of Display
     */
    public Display(int address, TdpDisplayProperties properties, TdpStatus status) {
        this.address = address;
        this.displayProperties = properties;
        this.status = status;

        timeoutSettings.setTimeoutMode(ETdpTimeoutMode.NO_TIMEOUT);
        diagnostics.setMessage("No diagnostics available");
        externalLighting.setState(ETdpLightingState.OFF);
    }

    @Override
    public void close() {
        showInfo.stopSlideShow();
    }

    /**
     * @return the address
     */
    public final int getAddress() {
        return address;
    }

    /**
     * @param address the address to set
     */
    public final void setAddress(int address) {
        this.address = address;
    }

    /**
     * @return the displayProperties
     */
    public final TdpDisplayProperties getDisplayProperties() {
        return displayProperties;
    }

    /**
     * @param displayProperties the displayProperties to set
     */
    public final void setDisplayProperties(TdpDisplayProperties displayProperties) {
        this.displayProperties = displayProperties;
    }

    /**
     * @return the status
     */
    public final TdpStatus getStatus() {
        return status;
    }

    /**
     * @param status the status to set
     */
    public final void setStatus(TdpStatus status) {
        this.status = status;
    }

    /**
     * @return the timeoutSettings
     */
    public final TdpTimeoutSettings getTimeoutSettings() {
        return timeoutSettings;
    }

    /**
     * @param timeoutSettings the timeoutSettings to set
     */
    public final void setTimeoutSettings(TdpTimeoutSettings timeoutSettings) {
        this.timeoutSettings = timeoutSettings;

        fireTimeoutChange();
    }

    /**
     * @return the brightnessTable
     */
    public TdpBrightnessTable getBrightnessTable() {
        return brightnessTable;
    }

    /**
     * @param brightnessTable the brightnessTable to set
     */
    public void setBrightnessTable(TdpBrightnessTable brightnessTable) {
        this.brightnessTable = brightnessTable;

        fireBrightnessChange();
    }

    /**
     * @return the diagnostics
     */
    public final TdpDiagnostics getDiagnostics() {
        return diagnostics;
    }

    /**
     * @param diagnostics the diagnostics to set
     */
    public final void setDiagnostics(TdpDiagnostics diagnostics) {
        this.diagnostics = diagnostics;
    }

    /**
     * @return the notifications
     */
    public final List<TdpNotification> getNotifications() {
        return notifications;
    }

    /**
     * @param notifications the notifications to set
     */
    public final void setNotifications(List<TdpNotification> notifications) {
        this.notifications = notifications;
    }

    public final void clearNotifications() {
        notifications = new ArrayList<TdpNotification>();
        fireClearNotificationsChange();
    }

    /**
     * @return the banks
     */
    public final ImageBanks getBanks() {
        return banks;
    }

    /**
     * @param banks the banks to set
     */
    public final void setBanks(ImageBanks banks) {
        this.banks = banks;
    }

    /**
     * @return the textLines
     */
    public final TdpSetText getTextLines() {
        return textLines;
    }

    /**
     * @param textLines the textLines to set
     */
    public final void setTextLines(TdpSetText textLines) {
        this.textLines = textLines;
    }

    /**
     * @return the showInfo
     */
    public final ShowInfo getShowInfo() {
        return showInfo;
    }

    /**
     * @param showInfo the showInfo to set
     */
    public final void setShowInfo(ShowInfo showInfo) {
        this.showInfo = showInfo;
    }

    public TdpExternalLighting getTdpExternalLighting() {
        return externalLighting;
    }

    public void setExternalLighting(TdpExternalLighting externalLighting) {
        this.externalLighting = externalLighting;

        fireExternalLightingChange();
    }

    /**
     * @param listener Listener of changes in external lighting
     */
    public final void addExternalLightingListener(IExternalLightingListener listener) {
        if (listener != null) {
            externalLightingListeners.add(listener);
        }
    }

    public final void removeExternalLightingListener(IExternalLightingListener listener) {
        if (listener != null) {
            externalLightingListeners.remove(listener);
        }
    }

    private void fireExternalLightingChange() {
        for (IExternalLightingListener listener : externalLightingListeners) {
            listener.externalLightingChanged(address, externalLighting);
        }
    }

    /**
     * @param listener Listener of changes in display
     */
    public final void addDisplayListener(IDisplayListener listener) {
        showInfo.addListener(listener);
    }

    public final void removeDisplayListener(IDisplayListener listener) {
        if (listener != null) {
            showInfo.removeListener(listener);
        }
    }

    /**
     * @param listener Listener of changes in settings
     */
    public final void addTimeoutListener(ICommunicationSettingsListener listener) {
        timeoutListeners.add(listener);
    }

    public final void removeTimeoutListener(ICommunicationSettingsListener listener) {
        if (listener != null) {
            timeoutListeners.remove(listener);
        }
    }

    private void fireTimeoutChange() {
        for (ICommunicationSettingsListener listener : timeoutListeners) {
            listener.settingsChanged(address, getTimeoutSettings());
        }
    }

    /**
     * @param listener Listener of changes in settings
     */
    public final void addBrightnessTableListener(IBrightnessTableListener listener) {
        brightnessListeners.add(listener);
    }

    public final void removeBrightnessTableListener(IBrightnessTableListener listener) {
        if (listener != null) {
            brightnessListeners.remove(listener);
        }
    }

    private void fireBrightnessChange() {
        for (IBrightnessTableListener listener : brightnessListeners) {
            listener.brightnessTableChanged(address, getBrightnessTable());
        }
    }

    /**
     * @param listener Listener of changes in settings
     */
    public final void addClearNotificationListener(IClearNotificationsListener listener) {
        clearListeners.add(listener);
    }

    public final void removeClearNotificationListener(IClearNotificationsListener listener) {
        if (listener != null) {
            clearListeners.remove(listener);
        }
    }

    private void fireClearNotificationsChange() {
        for (IClearNotificationsListener listener : clearListeners) {
            listener.clear(address);
        }
    }

}
