/*
Copyright (c) 2013-2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.displaysimulator.data;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.disperanto.communication.IStartableConnection;
import org.disperanto.displaysimulator.listeners.IAvailableDisplayListener;
import org.disperanto.messages.TdpMessage;
import org.disperanto.protocol.ITdpReceiveListener;
import org.disperanto.protocol.TdpProtocolReceiver;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Handles connection with server and handles incoming commands for the displays.
 */
public class DisplayController implements ITdpReceiveListener {
    private static final Logger LOG = LoggerFactory.getLogger(DisplayController.class);

    private final ILogger logger;

    private final List<IAvailableDisplayListener> availableDisplayListeners =
            new CopyOnWriteArrayList<IAvailableDisplayListener>();
    private final Map<Integer, Display> displays = new TreeMap<Integer, Display>();

    private IStartableConnection connection;
    private TdpProtocolReceiver receiver;

    /**
     * Constructor
     *
     * @param lm Model for logging panel
     */
    public DisplayController(ILogger lm) {
        logger = lm;
    }

    public void addDisplay(Display display) {
        if (displays.containsKey(display.getAddress())) {
            throw new IllegalArgumentException("Display was already added: " + display); //$NON-NLS-1$
        }

        displays.put(display.getAddress(), display);
        for (IAvailableDisplayListener ls : availableDisplayListeners) {
            ls.onDisplayAdded(display);
        }
    }

    public void removeDisplay(int address) {
        Display removed = displays.remove(address);
        if (removed != null) {
            for (IAvailableDisplayListener ls : availableDisplayListeners) {
                ls.onDisplayRemoved(removed);
            }
            removed.close();
        }
    }

    /**
     * @return List of displays
     */
    public final List<Display> getDisplays() {
        return new ArrayList<Display>(displays.values());
    }

    /**
     * Get display for specific address
     *
     * @param address Address of display
     * @return display
     */
    public final Display getDisplay(int address) {
        return displays.get(address);
    }

    @Override
    public final void commandReceived(List<TdpMessage> commands) {
        try {
            final List<TdpMessage> responses = MessageHandler.handleCommands(getDisplays(), commands, logger);

            if (responses != null && responses.size() > 0) {
                receiver.send(responses);
            }

        } catch (Exception e) { // SUPPRESS CHECKSTYLE IllegalCatchCheck
            logger.log("Exception: " + e.getMessage());
            LOG.warn("Exception receiving command", e);
        }
    }

    public IStartableConnection getConnection() {
        return connection;
    }

    /**
     * @param connection2 the connection to set
     */
    public final void setConnection(IStartableConnection connection2) {
        if (this.connection != null) {
            stopReceiving();
        }

        connection = connection2;

        startReceiving();

        logger.log(String.format("Connection settings are changed: %s://%s:%d",
                connection2.getConnectionType(), connection2.getIpAddress(), connection2.getPort()));
    }

    private void startReceiving() {
        try {
            connection.start();
        } catch (UnknownHostException e) {
            LOG.error("Unknown host", e);
            logger.log("Unknown host, cannot open connection");
        } catch (IOException e) {
            LOG.error("Unable to open socket", e);
            logger.log("Cannot open connection");
        }

        receiver = new TdpProtocolReceiver(connection);

        // add listener for incoming messages
        receiver.addReceiveListener(this);

        // start receiving
        final ExecutorService threadExecutor = Executors.newFixedThreadPool(1);
        threadExecutor.execute(receiver);
    }

    public void stopReceiving() {
        try {
            // No need to stop the connection first. stop() will handle that.
            receiver.stop();
        } catch (InterruptedException e) {
            LOG.error("Error trying to stop", e);
        }
    }

    public void addAvailableDisplayListener(IAvailableDisplayListener displayListener) {
        availableDisplayListeners.add(displayListener);
    }

    public void removeAvailableDisplayListener(IAvailableDisplayListener displayListener) {
        availableDisplayListeners.remove(displayListener);
    }

}
