/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.displaysimulator.ui.controllers;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.ArrayList;

import javax.swing.JCheckBox;
import javax.swing.JOptionPane;

import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.data.TdpStatus;
import org.disperanto.displaysimulator.data.DisplayController;
import org.disperanto.displaysimulator.listeners.IDisplayControllerListener;
import org.disperanto.displaysimulator.listeners.IPropertiesListener;
import org.disperanto.displaysimulator.ui.models.StatusModel;
import org.disperanto.displaysimulator.ui.panels.StatusPanel;
import org.disperanto.enums.ETdpLightingState;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * View controller for StatusPanel
 */
public class StatusViewController implements ActionListener, ItemListener, IPropertiesListener,
        IDisplayControllerListener {
    private static final Logger LOG = LoggerFactory.getLogger(StatusViewController.class);

    private StatusModel model;
    private StatusPanel panel;
    private DisplayController controller;
    private int displayAddress;

    /**
     * Constructor
     * 
     * @param panel Panel that is controlled
     * @param model Model of panel
     * @param controller Display
     */
    public StatusViewController(StatusPanel panel, StatusModel model, DisplayController controller) {
        this.model = model;
        this.panel = panel;
        this.controller = controller;
    }

    @Override
    public final void actionPerformed(ActionEvent a) {
        LOG.debug("Status changed");

        if (!validate()) {
            return;
        }

        setStatus();
    }

    @Override
    public final void itemStateChanged(ItemEvent e) {
        LOG.debug("Item changed");

        final Object source = e.getSource();
        if (source == panel.getLightSensorsCheck()) {
            final JCheckBox checkBox = (JCheckBox)source;
            if (checkBox.isSelected()) {

                panel.getMeasuredLightIntensity1Field().setEnabled(true);
                panel.getMeasuredLightIntensity2Field().setEnabled(true);
            } else {
                panel.getMeasuredLightIntensity1Field().setEnabled(false);
                panel.getMeasuredLightIntensity2Field().setEnabled(false);
            }
        }

        if (source == panel.getTemperatureCheck()) {
            final JCheckBox checkBox = (JCheckBox)source;
            if (checkBox.isSelected()) {
                panel.getTemperatureField().setEnabled(true);
            } else {
                panel.getTemperatureField().setEnabled(false);
            }
        }

        if (source == panel.getHeatingCheck()) {
            final JCheckBox checkBox = (JCheckBox)source;
            if (checkBox.isSelected()) {
                panel.getHeatingField().setEnabled(true);
            } else {
                panel.getHeatingField().setEnabled(false);
            }
        }

        if (source == panel.getCoolingCheck()) {
            final JCheckBox checkBox = (JCheckBox)source;
            if (checkBox.isSelected()) {
                panel.getCoolingField().setEnabled(true);
            } else {
                panel.getCoolingField().setEnabled(false);
            }
        }

    }

    private boolean validate() {

        if (!panel.getBrightnessField().isValueValid()) {
            JOptionPane.showMessageDialog(panel.getRootPane(),
                    "Brightness field is mandatory and should contain a valid between 0 and 100");
            return false;
        }

        if (model.getStatus().getIntensityExternalLighting() != null) {
            if (!panel.getIntensityExternalLightingField().isValueValid()) {
                JOptionPane.showMessageDialog(panel.getRootPane(),
                        "Intensity external lighting field is mandatory and should contain a valid between 0 and 100");
                return false;
            }
        }

        if (!panel.getIntensityExternalLightingField().isValueValid()) {
            JOptionPane.showMessageDialog(panel.getRootPane(),
                    "Intensity external lighting field is mandatory and should contain a valid between 0 and 100");
            return false;
        }

        if (panel.getLightSensorsCheck().isSelected()) {
            if (!panel.getMeasuredLightIntensity1Field().isValueValid() ||
                    !panel.getMeasuredLightIntensity2Field().isValueValid()) {
                JOptionPane.showMessageDialog(panel.getRootPane(), "Measured light intensity fields "
                        + "are mandatory when light sensors are present and should contain a valid between 0 and 100");
                return false;
            }
        }

        if (panel.getGpsPositionField().getText() == null) {
            JOptionPane.showMessageDialog(panel.getRootPane(), "GPS position field is mandatory and should"
                    + " contain a GPS position <longitude,latitude> in decimal values (e.g 5.659,51.473)");
            return false;
        }

        if (panel.getTemperatureCheck().isSelected()) {
            if (!panel.getTemperatureField().isValueValid()) {
                JOptionPane.showMessageDialog(panel.getRootPane(), "Temperature field is mandatory when "
                        + "temperature sensor is present and should contain a valid between 0 and 100");
                return false;
            }
        }

        if (panel.getHeatingCheck().isSelected()) {
            if (panel.getHeatingField().getSelectedItem() == null) {
                JOptionPane
                        .showMessageDialog(panel.getRootPane(), "Heating field is mandatory when heating is present");
                return false;
            }
        }

        if (panel.getCoolingCheck().isSelected()) {
            if (panel.getCoolingField().getSelectedItem() == null) {
                JOptionPane
                        .showMessageDialog(panel.getRootPane(), "Cooling field is mandatory when heating is present");
                return false;
            }
        }

        return true;
    }

    private void setStatus() {
        final TdpStatus status = model.getStatus();

        status.setBrightnessInPercentage((int)panel.getBrightnessField().getValue());
        if (panel.getIntensityExternalLightingField().isEnabled()) {
            status.setIntensityExternalLighting((int)panel.getIntensityExternalLightingField().getValue());
        }
        if (panel.getLightSensorsCheck().isSelected()) {
            final ArrayList<Integer> list = new ArrayList<Integer>();
            list.add((int)panel.getMeasuredLightIntensity1Field().getValue());
            list.add((int)panel.getMeasuredLightIntensity2Field().getValue());

            status.setMeasuredLightIntensity(list);
        } else {
            status.setMeasuredLightIntensity(new ArrayList<Integer>()); // empty list
        }
        if (panel.getTemperatureCheck().isSelected()) {
            status.setTemperature((int)panel.getTemperatureField().getValue());
        } else {
            status.setTemperature(null);
        }
        if (panel.getHeatingCheck().isSelected()) {
            ETdpLightingState state = (ETdpLightingState)panel.getHeatingField().getSelectedItem();
            status.setHeating(state.getValue());
        } else {
            status.setHeating(null);
        }
        if (panel.getCoolingCheck().isSelected()) {
            final ETdpLightingState state = (ETdpLightingState)panel.getCoolingField().getSelectedItem();
            status.setCooling(state.getValue());
        } else {
            status.setCooling(null);
        }

        model.setStatus(status);

        // copy object to display
        controller.getDisplay(displayAddress).setStatus(new TdpStatus(status));
    }

    @Override
    public final void propertiesChanged(int displayAddress, TdpDisplayProperties properties) {
        final TdpStatus s = model.getStatus();

        LOG.debug("Status: display properties changed");
        if (properties.hasExternalLighting()) {
            if (s.getIntensityExternalLighting() == null) {
                s.setIntensityExternalLighting(100);
                model.setStatus(s);
            }
        } else {
            if (s.getIntensityExternalLighting() != null) {
                s.setIntensityExternalLighting(null);
                model.setStatus(s);
            }
        }
    }

    @Override
    public final void displayChanged(int activeDisplay) {
        displayAddress = activeDisplay;

        // init model: copy initial values from display
        model.setStatus(new TdpStatus(controller.getDisplay(displayAddress).getStatus()));
    }
}
