/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.simulator;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;

import org.disperanto.simulator.controllers.CommandChoiceViewController;
import org.disperanto.simulator.controllers.CommandChoiceViewModel;
import org.disperanto.simulator.controllers.CommandListViewController;
import org.disperanto.simulator.controllers.ConnectionViewController;
import org.disperanto.simulator.controllers.ResponseListViewController;
import org.disperanto.simulator.controllers.Server;
import org.disperanto.simulator.data.LogEvent;
import org.disperanto.simulator.ui.models.LoggingModel;
import org.disperanto.simulator.ui.panels.BrightnessTablePanel;
import org.disperanto.simulator.ui.panels.ConnectionPanel;
import org.disperanto.simulator.ui.panels.CrcListPanel;
import org.disperanto.simulator.ui.panels.DataCommunicationTimeoutPanel;
import org.disperanto.simulator.ui.panels.ExchangePanel;
import org.disperanto.simulator.ui.panels.ExternalLightingPanel;
import org.disperanto.simulator.ui.panels.LoggingPanel;
import org.disperanto.simulator.ui.panels.ManipulateMemorySlotPanel;
import org.disperanto.simulator.ui.panels.SetTextPanel;
import org.disperanto.simulator.ui.panels.ShowImagePanel;
import org.disperanto.simulator.ui.panels.SlideShowPanel;

import net.miginfocom.swing.MigLayout;

/**
 * Main dialog for Server simulator for Disperanto.
 */
public class SimulatorMainDialog extends JFrame {
    private static final String VERSION = "v1.2";

    private final Server server;

    private DataCommunicationTimeoutPanel timeoutPanel;
    private BrightnessTablePanel brightnessTablePanel;
    private ExternalLightingPanel externalLightingPanel;
    private CrcListPanel crcPanel;
    private ShowImagePanel showImagePanel;
    private SlideShowPanel slideShowPanel;
    private ManipulateMemorySlotPanel manipulatePanel;
    private SetTextPanel setTextPanel;

    private final LoggingModel loggingModel;

    /**
     * Constructor
     *
     * @param title Application title
     * @param server server for TDP protocol
     * @param loggingModel Model for logging panel
     */
    public SimulatorMainDialog(String title, Server server, LoggingModel loggingModel) {
        super(title);
        loggingModel.addRow(new LogEvent(title + " " + VERSION));

        this.server = server;
        this.loggingModel = loggingModel;

        addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent we) {
                System.exit(0);
                super.windowClosing(we);
            }
        });

        final JPanel panel = new JPanel();
        panel.setLayout(new BorderLayout(10, 10));
        panel.add(createMainPanel(), BorderLayout.CENTER);
        setContentPane(panel);

        pack();
        setSize(new Dimension(1280, 768));
        setLocationRelativeTo(null);
    }

    private Component createMainPanel() {
        // Panel for communication and handling messages
        final JPanel panelMessages = new JPanel(new MigLayout("", "[grow]", "[grow]"));

        // Panel for message data
        final JPanel panelData = createMessageDataPanel();

        panelMessages.add(createConnectionPanel(), "north");
        panelMessages.add(new LoggingPanel(loggingModel), "south");
        panelMessages.add(createExchangePanel(), "grow");

        // Add panels to tabbed panel
        JTabbedPane tabbedPane = new JTabbedPane();
        tabbedPane.addTab("Messages", panelMessages);
        tabbedPane.addTab("Data", panelData);

        return tabbedPane;
    }

    private JPanel createConnectionPanel() {
        ConnectionPanel connectionPanel = new ConnectionPanel();
        ConnectionViewController controller = new ConnectionViewController(connectionPanel, server, loggingModel);
        connectionPanel.addListener(controller);

        return connectionPanel;
    }

    private JPanel createExchangePanel() {
        ExchangePanel exchangePanel = new ExchangePanel();

        CommandChoiceViewModel choiceModel = new CommandChoiceViewModel(timeoutPanel, brightnessTablePanel,
                externalLightingPanel, crcPanel, showImagePanel, slideShowPanel, manipulatePanel, setTextPanel);

        CommandChoiceViewController choiceController = new CommandChoiceViewController(
                exchangePanel.getCommandChoicePanel(), server, choiceModel);
        exchangePanel.getCommandChoicePanel().addListener(choiceController);
        choiceController.addListener(exchangePanel.getCommandListPanel());

        CommandListViewController listController = new CommandListViewController(exchangePanel.getCommandListPanel(),
                server, loggingModel);
        listController.addListener(exchangePanel.getResponseListPanel());

        ResponseListViewController responseController = 
                new ResponseListViewController(exchangePanel.getResponseListPanel());

        exchangePanel.getCommandListPanel().addListener(listController);
        exchangePanel.getCommandListPanel().getCommandTree().addTreeSelectionListener(listController);
        exchangePanel.getResponseListPanel().getResponseTree().addTreeSelectionListener(responseController);

        server.addNotificationListener(exchangePanel.getResponseListPanel());

        return exchangePanel;
    }

    private JPanel createMessageDataPanel() {
        JPanel dataPanel = new JPanel(new MigLayout("", "[]40[]40[]", ""));

        timeoutPanel = new DataCommunicationTimeoutPanel();
        server.addCommandListChangeListener(timeoutPanel);
        dataPanel.add(timeoutPanel, "grow");
        
        manipulatePanel = new ManipulateMemorySlotPanel();
        server.addCommandListChangeListener(manipulatePanel);
        dataPanel.add(manipulatePanel, "spany 5");

        slideShowPanel = new SlideShowPanel();
        server.addCommandListChangeListener(slideShowPanel);
        dataPanel.add(slideShowPanel, "spany 6, wrap");
        
        externalLightingPanel = new ExternalLightingPanel();
        server.addCommandListChangeListener(externalLightingPanel);
        dataPanel.add(externalLightingPanel, "grow,wrap");

        crcPanel = new CrcListPanel();
        server.addCommandListChangeListener(crcPanel);
        dataPanel.add(crcPanel, "grow,wrap");

        showImagePanel = new ShowImagePanel();
        server.addCommandListChangeListener(showImagePanel);
        dataPanel.add(showImagePanel, "top, growx, wrap");

        setTextPanel = new SetTextPanel();
        server.addCommandListChangeListener(setTextPanel);
        dataPanel.add(setTextPanel, "top, growx, wrap");

        brightnessTablePanel = new BrightnessTablePanel();
        server.addCommandListChangeListener(brightnessTablePanel);
        dataPanel.add(brightnessTablePanel, "spanx 2");

        return dataPanel;
    }
}
