/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.simulator.controllers;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.swing.JOptionPane;

import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpNotification;
import org.disperanto.messages.TdpCommandFactory;
import org.disperanto.messages.TdpMessage;
import org.disperanto.simulator.listeners.ICommandChoiceListener;
import org.disperanto.simulator.ui.models.IBrightnessTableModel;
import org.disperanto.simulator.ui.models.ICommunicationTimeoutSettingsModel;
import org.disperanto.simulator.ui.models.ICrcListModel;
import org.disperanto.simulator.ui.models.IExternalLightingModel;
import org.disperanto.simulator.ui.models.IManipulateMemorySlotModel;
import org.disperanto.simulator.ui.models.ISetTextModel;
import org.disperanto.simulator.ui.models.IShowImageModel;
import org.disperanto.simulator.ui.models.ISlideShowModel;
import org.disperanto.simulator.ui.panels.CommandChoicePanel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * View controller for CommandChoicePanel
 */
public class CommandChoiceViewController implements ActionListener {
    private static final Logger LOG = LoggerFactory.getLogger(CommandChoiceViewController.class);

    private final Server server;
    private final CommandChoicePanel panel;

    private final ICommunicationTimeoutSettingsModel timeoutModel;
    private final IBrightnessTableModel brightnessTableModel;
    private final IExternalLightingModel lightingModel;
    private final ICrcListModel crcModel;
    private final IShowImageModel showImageModel;
    private final ISlideShowModel slideShowModel;
    private final IManipulateMemorySlotModel manipulateMemorySlotModel;
    private final ISetTextModel setTextModel;

    private final List<ICommandChoiceListener> listeners = new ArrayList<ICommandChoiceListener>();

    public CommandChoiceViewController(CommandChoicePanel panel, Server server, CommandChoiceViewModel model) {
        this.panel = panel;
        this.server = server;

        this.timeoutModel = model.getTimeoutModel();
        this.brightnessTableModel = model.getBrightnessTableModel();
        this.lightingModel = model.getLightingModel();
        this.crcModel = model.getCrcModel();
        this.showImageModel = model.getShowImageModel();
        this.slideShowModel = model.getSlideShowModel();
        this.manipulateMemorySlotModel = model.getManipulateMemorySlotModel();
        this.setTextModel = model.getSetTextModel();
    }

    /**
     * @param listener Listener of changes in command list
     */
    public final void addListener(ICommandChoiceListener listener) {
        listeners.add(listener);
    }

    @Override
    public final void actionPerformed(ActionEvent arg0) {
        LOG.debug("Add button pressed");

        final List<ETdpCommand> commandIdList = fillCommandIdList();

        if (!validate(commandIdList)) {
            return;
        }

        addCommands(commandIdList);

        resetPanelFields();
    }

    private boolean validate(List<ETdpCommand> commandIdList) {
        if (commandIdList.isEmpty()) {
            JOptionPane.showMessageDialog(panel, "Select at least one message");
            return false;
        }

        for (ETdpCommand commandId : commandIdList) {
            if (server.getCommandList().containsKey(commandId)) {
                if (server.getCommandList().get(commandId).getAddresses().contains(panel.getDisplayAddress())) {
                    JOptionPane.showMessageDialog(panel,
                            "List of commands to send already contains a " + commandId.getName() +
                                    " for display address " + panel.getDisplayAddress());
                    return false;
                }
            }
        }

        return true;
    }

    // CHECKSTYLE:OFF (Cyclomatic complexity too high)
    private void addCommands(List<ETdpCommand> commandIdList) {
        // CHECKSTYLE: ON
        final HashMap<ETdpCommand, TdpMessage> commands = new HashMap<ETdpCommand, TdpMessage>();

        final int address = panel.getDisplayAddress();

        for (ETdpCommand commandId : commandIdList) {
            // get default message for command id
            switch (commandId) {
            case DISPLAY_PROPERTIES:
                commands.put(ETdpCommand.DISPLAY_PROPERTIES, createDisplayPropertiesMessage(address));
                break;
            case STATUS:
                commands.put(ETdpCommand.STATUS, createStatusMessage(address));
                break;
            case REBOOT:
                commands.put(ETdpCommand.REBOOT, createRebootMessage(address));
                break;
            case KEEP_ALIVE:
                commands.put(ETdpCommand.KEEP_ALIVE, createKeepAliveMessage(address));
                break;
            case SET_COMMUNICATION_TIMEOUT:
                commands.put(ETdpCommand.SET_COMMUNICATION_TIMEOUT, createSetCommunicationTimeoutMessage(address));
                break;
            case SET_BRIGHTNESS:
                commands.put(ETdpCommand.SET_BRIGHTNESS, createSetBrightnessTableMessage(address));
                break;
            case SET_EXTERNAL_LIGHTING:
                commands.put(ETdpCommand.SET_EXTERNAL_LIGHTING, createSetExternalLightingMessage(address));
                break;
            case DIAGNOSTICS:
                commands.put(ETdpCommand.DIAGNOSTICS, createDiagnosticsMessage(address));
                break;
            case MANIPULATE_MEMORY_SLOT:
                commands.put(ETdpCommand.MANIPULATE_MEMORY_SLOT, createManipulateMemorySlotMessage(address));
                break;
            case CALCULATE_CRC_IMAGES:
                commands.put(ETdpCommand.CALCULATE_CRC_IMAGES, createCalculateCrcMessage(address));
                break;
            case SHOW_NO_IMAGE:
                commands.put(ETdpCommand.SHOW_NO_IMAGE, createShowNoImageMessage(address));
                break;
            case SHOW_IMAGE:
                commands.put(ETdpCommand.SHOW_IMAGE, createShowImageMessage(address));
                break;
            case START_SLIDE_SHOW:
                commands.put(ETdpCommand.START_SLIDE_SHOW, createStartSlideShowMessage(address));
                break;
            case SET_TEXT:
                commands.put(ETdpCommand.SET_TEXT, createSetTextMessage(address));
                break;
            case SUPPLIER_SERVICE_MODE:
                commands.put(ETdpCommand.SUPPLIER_SERVICE_MODE, createServiceModeMessage(address));
                break;
            case NOTIFICATION:
                commands.put(ETdpCommand.NOTIFICATION, createClearNotificationsMessage(address));
                break;
            default:
                throw new IllegalArgumentException("Unexpected commandId received: " + commandId.getName());
            }
        }

        server.addCommands(commands, address);
        fireChange();
    }

    private List<ETdpCommand> fillCommandIdList() {
        final List<ETdpCommand> list = new ArrayList<ETdpCommand>();

        if (panel.getDisplayPropertiesCheck().isSelected()) {
            list.add(ETdpCommand.DISPLAY_PROPERTIES);
        }
        if (panel.getStatusCheck().isSelected()) {
            list.add(ETdpCommand.STATUS);
        }
        if (panel.getRebootCheck().isSelected()) {
            list.add(ETdpCommand.REBOOT);
        }
        if (panel.getKeepAliveCheck().isSelected()) {
            list.add(ETdpCommand.KEEP_ALIVE);
        }
        if (panel.getSetCommunicationTimeoutCheck().isSelected()) {
            list.add(ETdpCommand.SET_COMMUNICATION_TIMEOUT);
        }
        if (panel.getSetBrightnessTableCheck().isSelected()) {
            list.add(ETdpCommand.SET_BRIGHTNESS);
        }
        if (panel.getSetExternalLightingCheck().isSelected()) {
            list.add(ETdpCommand.SET_EXTERNAL_LIGHTING);
        }
        if (panel.getDiagnosticsCheck().isSelected()) {
            list.add(ETdpCommand.DIAGNOSTICS);
        }
        if (panel.getManipulateMemorySlotCheck().isSelected()) {
            list.add(ETdpCommand.MANIPULATE_MEMORY_SLOT);
        }
        if (panel.getCalculateCrcCheck().isSelected()) {
            list.add(ETdpCommand.CALCULATE_CRC_IMAGES);
        }
        if (panel.getShowNoImageCheck().isSelected()) {
            list.add(ETdpCommand.SHOW_NO_IMAGE);
        }
        if (panel.getShowImageCheck().isSelected()) {
            list.add(ETdpCommand.SHOW_IMAGE);
        }
        if (panel.getStartSlideShowCheck().isSelected()) {
            list.add(ETdpCommand.START_SLIDE_SHOW);
        }
        if (panel.getSetTextCheck().isSelected()) {
            list.add(ETdpCommand.SET_TEXT);
        }
        if (panel.getServiceModeCheck().isSelected()) {
            list.add(ETdpCommand.SUPPLIER_SERVICE_MODE);
        }
        if (panel.getClearNotificationsCheck().isSelected()) {
            list.add(ETdpCommand.NOTIFICATION);
        }

        return list;
    }

    private void resetPanelFields() {
        panel.getDisplayPropertiesCheck().setSelected(false);
        panel.getStatusCheck().setSelected(false);
        panel.getRebootCheck().setSelected(false);
        panel.getKeepAliveCheck().setSelected(false);
        panel.getSetCommunicationTimeoutCheck().setSelected(false);
        panel.getSetBrightnessTableCheck().setSelected(false);
        panel.getSetExternalLightingCheck().setSelected(false);
        panel.getDiagnosticsCheck().setSelected(false);
        panel.getManipulateMemorySlotCheck().setSelected(false);
        panel.getCalculateCrcCheck().setSelected(false);
        panel.getShowNoImageCheck().setSelected(false);
        panel.getShowImageCheck().setSelected(false);
        panel.getStartSlideShowCheck().setSelected(false);
        panel.getSetTextCheck().setSelected(false);
        panel.getServiceModeCheck().setSelected(false);
        panel.getClearNotificationsCheck().setSelected(false);
    }

    /**
     * fires notification for listeners.
     */
    public final void fireChange() {
        LOG.debug("Fire change");

        final List<TdpMessage> list = new ArrayList<TdpMessage>();
        list.addAll(server.getCommandList().values());

        for (ICommandChoiceListener listener : listeners) {
            listener.commandListChanged(list);
        }
    }

    private TdpMessage createDisplayPropertiesMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createDisplayPropertiesCommand(list);
    }

    private TdpMessage createStatusMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createStatusCommand(list);
    }

    private TdpMessage createRebootMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createRebootCommand(list);
    }

    private TdpMessage createKeepAliveMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createKeepAliveCommand(list);
    }

    private TdpMessage createSetCommunicationTimeoutMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createSetCommunicationTimeoutCommand(list, timeoutModel.getTimeoutSettings());
    }

    private TdpMessage createSetBrightnessTableMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createSetBrightnessTableCommand(list, brightnessTableModel.getBrightnessTable());
    }

    private TdpMessage createSetExternalLightingMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createSetExternalLightingCommand(list, lightingModel.getLightingSettings());
    }

    private TdpMessage createDiagnosticsMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createDiagnosticsCommand(list);
    }

    private TdpMessage createManipulateMemorySlotMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createManipulateMemorySlotCommand(list, 
                manipulateMemorySlotModel.getManipulateMemorySlotSettings());
    }

    private TdpMessage createCalculateCrcMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createCalculateCrcImagesCommand(list, crcModel.getSlotList());
    }

    private TdpMessage createShowNoImageMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createShowNoImageCommand(list);
    }

    private TdpMessage createShowImageMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createShowImageCommand(list, showImageModel.getShowImage());
    }

    private TdpMessage createStartSlideShowMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createStartSlideShowCommand(list, slideShowModel.getSlideShowInfo());
    }

    private TdpMessage createSetTextMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createSetTextCommand(list, setTextModel.getSetText());
    }

    private TdpMessage createServiceModeMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        return TdpCommandFactory.createServiceModeCommand(list);
    }

    private TdpMessage createClearNotificationsMessage(int address) {
        final List<Integer> list = new ArrayList<Integer>();
        list.add(address);

        // just clear all
        List<ETdpNotification> toClear = new ArrayList<ETdpNotification>();
        for (ETdpNotification notification : ETdpNotification.values()) {
            if (notification.isNeedsConfirm()) {
                toClear.add(notification);
            }
        }

        return TdpCommandFactory.createClearNotificationsCommand(list, toClear);
    }

}
