/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.simulator.controllers;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.JOptionPane;
import javax.swing.JTree;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;

import org.disperanto.messages.TdpMessage;
import org.disperanto.simulator.data.LogEvent;
import org.disperanto.simulator.listeners.IResponseListener;
import org.disperanto.simulator.ui.models.LoggingModel;
import org.disperanto.simulator.ui.panels.CommandListPanel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * View controller for CommandListPanel
 */
public class CommandListViewController implements ActionListener, TreeSelectionListener {
    private static final Logger LOG = LoggerFactory.getLogger(CommandListViewController.class);

    private final Server server;
    private final CommandListPanel panel;
    private final LoggingModel loggingModel;

    private final List<IResponseListener> listeners = new ArrayList<IResponseListener>();

    /**
     * Constructor
     *
     * @param panel Panel that is controlled
     * @param server server
     * @param lm Model for logging panel
     */
    public CommandListViewController(CommandListPanel panel, Server server, LoggingModel lm) {
        this.panel = panel;
        this.server = server;
        loggingModel = lm;
    }

    /**
     * @param listener Listener of changes in response list
     */
    public final void addListener(IResponseListener listener) {
        listeners.add(listener);
    }

    @Override
    public final void actionPerformed(ActionEvent ae) {
        if (ae.getSource() == panel.getSendButton()) {
            handleSendCommands();
        } else if (ae.getSource() == panel.getRemoveButton()) {
            handleRemoveSelected();
        } else if (ae.getSource() == panel.getClearAllButton()) {
            handleClearAll();
        } else {
            throw new IllegalArgumentException("Unknown source for panel action");
        }
    }

    private void handleRemoveSelected() {
        LOG.debug("Remove button pressed");

        final JTree commandTree = panel.getCommandTree();
        final DefaultMutableTreeNode node = (DefaultMutableTreeNode)commandTree.getLastSelectedPathComponent();

        if (node == null) {
            return;
        } else {
            if (!node.isRoot()) {
                updateCommands(node);

                // update tree model
                final DefaultTreeModel model = (DefaultTreeModel)commandTree.getModel();
                final DefaultMutableTreeNode parent = (DefaultMutableTreeNode)node.getParent();
                if (parent.getChildCount() > 1) {
                    model.removeNodeFromParent(node);
                } else {
                    if (parent.isRoot()) {
                        model.removeNodeFromParent(node);
                    } else {
                        model.removeNodeFromParent(parent);
                    }
                }

            }

            commandTree.updateUI();
        }
    }

    private void handleClearAll() {
        LOG.debug("Clear all button pressed");

        final JTree commandTree = panel.getCommandTree();
        final DefaultMutableTreeNode root = (DefaultMutableTreeNode)commandTree.getModel().getRoot();

        root.removeAllChildren();

        commandTree.updateUI();

        panel.getMessageData().setText("");

        server.clearAllCommands();

        fireClearResponses();
    }

    private void handleSendCommands() {
        LOG.debug("Send button pressed");

        if (!validate()) {
            return;
        }

        sendCommands();
    }

    private boolean validate() {
        final JTree commandTree = panel.getCommandTree();
        final DefaultMutableTreeNode root = (DefaultMutableTreeNode)commandTree.getModel().getRoot();

        if (root.getChildCount() == 0) {
            JOptionPane.showMessageDialog(panel.getRootPane(), "Add at least one message");
            return false;
        }

        return true;
    }

    private void sendCommands() {
        loggingModel.addRow(new LogEvent(server.getCommandList().size() + " command(s) send to display controller"));
        final List<TdpMessage> responses = server.exchange(new ArrayList<TdpMessage>(server.getCommandList().values()));

        if (responses == null || responses.size() == 0) {
            loggingModel.addRow(new LogEvent("No responses received from display controller"));
        } else {
            loggingModel.addRow(new LogEvent(responses.size() + " response(s) received from display controller"));
            fireResponseReceived(responses);
        }
    }

    private void updateCommands(DefaultMutableTreeNode node) {
        if (node.getUserObject() instanceof TdpMessage) {
            // command selected
            final TdpMessage message = (TdpMessage)node.getUserObject();
            server.removeCommand(message.getCommand());
        } else if (node.getUserObject() instanceof Integer) {
            // address selected
            final DefaultMutableTreeNode parent = (DefaultMutableTreeNode)node.getParent();
            final TdpMessage message = (TdpMessage)parent.getUserObject();
            if (message.getAddresses().size() == 1) {
                // remove command
                server.removeCommand(message.getCommand());
            } else {
                // remove address from command
                server.removeAddressFromCommand(message.getCommand(), (Integer)node.getUserObject());
            }
        }

        LOG.debug("updateCommands");
    }

    private void fireResponseReceived(List<TdpMessage> responses) {
        LOG.debug("Fire change");

        for (IResponseListener listener : listeners) {
            listener.responseListChanged(responses);
        }
    }

    private void fireClearResponses() {
        LOG.debug("Fire clear responses");

        for (IResponseListener listener : listeners) {
            listener.clearResponses();
        }
    }

    @Override
    public void valueChanged(TreeSelectionEvent arg0) {
        DefaultMutableTreeNode node = (DefaultMutableTreeNode)panel.getCommandTree().getLastSelectedPathComponent();
        TdpMessage command;

        /* if nothing is selected */
        if (node == null) {
            return;
        }

        /* retrieve the node that was selected */
        if (node.getUserObject() instanceof TdpMessage) {
            command = (TdpMessage)node.getUserObject();
        } else if (node.getUserObject() instanceof Integer) {
            final DefaultMutableTreeNode parent = (DefaultMutableTreeNode)node.getParent();
            command = (TdpMessage)parent.getUserObject();
        } else {
            return;
        }

        panel.getMessageData().setText(command.toReadableString());
        panel.getMessageData().setCaretPosition(0);
    }
}
