/*
Copyright (c) 2013-2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.simulator.controllers;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.disperanto.communication.IStartableConnection;
import org.disperanto.data.TdpNotifications;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.messages.TdpMessage;
import org.disperanto.protocol.ITdpNotificationListener;
import org.disperanto.protocol.TdpProtocolExchanger;
import org.disperanto.simulator.data.LogEvent;
import org.disperanto.simulator.listeners.ICommandListChangeListener;
import org.disperanto.simulator.listeners.INotificationListener;
import org.disperanto.simulator.ui.models.LoggingModel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Handles connection with display controller and sends commands and handles incoming responses for the displays.
 */
public class Server implements ITdpNotificationListener {

    private static final Logger LOG = LoggerFactory.getLogger(Server.class);

    private final LoggingModel loggingModel;

    private IStartableConnection connection;
    private TdpProtocolExchanger exchanger;

    private final HashMap<ETdpCommand, TdpMessage> commandList = new HashMap<ETdpCommand, TdpMessage>();
    private final List<INotificationListener> notificationListeners = new ArrayList<INotificationListener>();
    private final List<ICommandListChangeListener> commandListChangeListeners = 
            new ArrayList<ICommandListChangeListener>();

    /**
     * Constructor
     *
     * @param lm Model for logging panel
     */
    public Server(LoggingModel lm) {
        loggingModel = lm;
    }

    /**
     * Sends messages to display controller and returns the responses.
     *
     * @param commands List of commands that are send to display controller
     * @return List of responses from display controller
     */
    public final List<TdpMessage> exchange(List<TdpMessage> commands) {
        try {
            return exchanger.exchangeMessages(commands);
        } catch (IOException | TdpProtocolException e) {
            loggingModel.addRow(new LogEvent(e.toString()));
        }

        // An error occurred, reset the connection
        try {
            restartConnection();
        } catch (IOException ioe) {
            loggingModel.addRow(new LogEvent("Error restarting connection: " + ioe.toString()));
        }
        return null;
    }

    private void restartConnection() throws IOException {
        if (connection != null) {
            connection.stop();
            connection.start();
        }
    }

    /**
     * @return the connection
     */
    public final IStartableConnection getConnection() {
        return connection;
    }

    /**
     * @param connection2 the connection to set
     */
    public final void setConnection(IStartableConnection connection2) {
        if (this.connection != null) {
            this.connection.stop();
        }

        if (exchanger != null) {
            exchanger.removeNotificationListener(this);
        }

        this.connection = connection2;

        try {
            connection2.start();
        } catch (UnknownHostException e) {
            LOG.error("Unknown host", e);
            loggingModel.addRow(new LogEvent("Unknown host, cannot open connection"));
        } catch (IOException e) {
            LOG.error("Unable to open socket", e);
            loggingModel.addRow(new LogEvent("Cannot open connection"));
        }

        exchanger = new TdpProtocolExchanger(connection2);
        exchanger.addNotificationListener(this);

        loggingModel.addRow(new LogEvent(String.format("Connection settings are changed: %s://%s:%d",
                connection2.getConnectionType(), connection2.getIpAddress(), connection2.getPort())));
    }

    /**
     * @return the commandList
     */
    public final HashMap<ETdpCommand, TdpMessage> getCommandList() {
        return commandList;
    }

    public final void addCommands(Map<ETdpCommand, TdpMessage> commands, int address) {
        for (Entry<ETdpCommand, TdpMessage> entry : commands.entrySet()) {
            ETdpCommand command = entry.getKey();
            TdpMessage message = entry.getValue();

            if (commandList.get(command) != null &&
                    !commandList.get(command).getAddresses().contains(Integer.valueOf(address))) {
                commandList.get(command).addAddress(Integer.valueOf(address));
            } else {
                commandList.put(command, message);
            }
        }
        commandListChanged();
    }

    public final void clearAllCommands() {
        commandList.clear();
        commandListChanged();
    }

    public final void removeCommand(ETdpCommand command) {
        commandList.remove(command);
        commandListChanged();
    }

    public final void removeAddressFromCommand(ETdpCommand command, Integer address) {
        commandList.get(command).getAddresses().remove(address);
        commandListChanged();
    }

    /**
     * Adds a listener for received commands.
     *
     * @param listener Listener for received commands
     */
    public final void addNotificationListener(INotificationListener listener) {
        notificationListeners.add(listener);
    }

    public final void addCommandListChangeListener(ICommandListChangeListener listener) {
        commandListChangeListeners.add(listener);
    }

    /**
     * Notifies a notification to the listeners
     *
     * @throws TdpProtocolException
     * @throws IOException
     */
    private void notify(TdpNotifications notifications) throws TdpProtocolException, IOException {
        for (INotificationListener listener : notificationListeners) {
            listener.notificationReceived(notifications);
        }
    }

    @Override
    public final void notificationReceived(TdpNotifications notifications) throws TdpProtocolException, IOException {
        loggingModel.addRow(new LogEvent(notifications.getNotifications().size() + " notifications received"));

        notify(notifications);
    }

    private void commandListChanged() {
        for (ICommandListChangeListener listener : commandListChangeListeners) {
            listener.commandListChanged(commandList.keySet());
        }
    }
}
