/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.simulator.ui.panels;

import java.awt.Font;
import java.awt.event.ActionListener;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeSelectionModel;

import net.miginfocom.swing.MigLayout;

import org.disperanto.messages.TdpMessage;
import org.disperanto.simulator.listeners.ICommandChoiceListener;
import org.disperanto.simulator.utils.JTreeUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Panel for showing the commands to send.
 */
public class CommandListPanel extends JPanel implements ICommandChoiceListener {
    private static final Logger LOG = LoggerFactory.getLogger(CommandListPanel.class);

    // Components
    private JTree commandTree;

    private JButton sendButton;
    private JButton removeButton;
    private JButton clearAllButton;

    private JTextArea messageData;

    /**
     * Constructor
     */
    public CommandListPanel() {
        initUI();
    }

    /**
     * Add listener for panel actions
     *
     * @param listener Action listener
     */
    public final void addListener(ActionListener listener) {
        sendButton.addActionListener(listener);
        removeButton.addActionListener(listener);
        clearAllButton.addActionListener(listener);
    }

    private void initUI() {
        setLayout(new MigLayout("", "[grow][]", ""));
        setBorder(BorderFactory.createTitledBorder("Command list"));

        final DefaultMutableTreeNode top = new DefaultMutableTreeNode("Command list");
        DefaultTreeModel treeModel = new DefaultTreeModel(top);

        commandTree = new JTree(treeModel) {
            @Override
            public String convertValueToText(Object value, boolean selected, boolean expanded, boolean leaf, int row,
                    boolean hasFocus) {
                final DefaultMutableTreeNode node = (DefaultMutableTreeNode)value;
                if (node.getUserObject() instanceof TdpMessage) {
                    final TdpMessage message = (TdpMessage)node.getUserObject();
                    return super.convertValueToText(message.getCommand().getName(), selected, expanded, leaf, row,
                            hasFocus);
                } else {
                    return super.convertValueToText(value, selected, expanded, leaf, row, hasFocus);
                }
            }
        };
        commandTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);

        final JScrollPane treeView = new JScrollPane(commandTree);
        add(treeView, "span 1 3, grow");

        removeButton = new JButton("Remove command");
        add(removeButton, "wrap, growx");

        clearAllButton = new JButton("Clear all commands");
        add(clearAllButton, "wrap, growx");

        sendButton = new JButton("Send commands");
        add(sendButton, "wrap, growx, bottom");

        add(new JLabel("Message data:"), "wrap");
        messageData = new JTextArea(10, 30);
        messageData.setFont(new Font("Tahoma", 0, 11));
        messageData.setEditable(false);
        final JScrollPane scrollPane = new JScrollPane(messageData);
        add(scrollPane, "span 2, wrap, grow");
    }

    @Override
    public final void commandListChanged(List<TdpMessage> commandList) {
        LOG.debug("Build command tree");
        final DefaultMutableTreeNode top = (DefaultMutableTreeNode)commandTree.getModel().getRoot();

        top.removeAllChildren();

        for (TdpMessage command : commandList) {
            final DefaultMutableTreeNode node = new DefaultMutableTreeNode(command);
            top.add(node);

            for (Integer address : command.getAddresses()) {
                node.add(new DefaultMutableTreeNode(address));
            }
        }

        JTreeUtils.expandAll(commandTree, true);

        commandTree.updateUI();
        messageData.setText("");

        return;
    }

    /**
     * @return the commandTree
     */
    public final JTree getCommandTree() {
        return commandTree;
    }

    /**
     * @return the messageData
     */
    public final JTextArea getMessageData() {
        return messageData;
    }

    /**
     * @return the sendButton
     */
    public final JButton getSendButton() {
        return sendButton;
    }

    /**
     * @return the removeButton
     */
    public final JButton getRemoveButton() {
        return removeButton;
    }

    /**
     * @return the clearAllButton
     */
    public final JButton getClearAllButton() {
        return clearAllButton;
    }
}
