/*
Copyright (c) 2013-2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.simulator.ui.panels;

import java.awt.Font;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeSelectionModel;

import org.disperanto.data.TdpCommunicationError;
import org.disperanto.data.TdpNotification;
import org.disperanto.data.TdpNotifications;
import org.disperanto.messages.TdpMessage;
import org.disperanto.simulator.listeners.INotificationListener;
import org.disperanto.simulator.listeners.IResponseListener;
import org.disperanto.simulator.utils.JTreeUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import net.miginfocom.swing.MigLayout;

/**
 * Panel showing the received responses.
 */
public class ResponseListPanel extends JPanel implements IResponseListener, INotificationListener {
    private static final Logger LOG = LoggerFactory.getLogger(ResponseListPanel.class);

    // Components
    private JTree responseTree;
    private JTree notificationTree;
    private JTextArea messageData;

    /**
     * Constructor
     */
    public ResponseListPanel() {
        initUI();
    }

    private void initUI() {
        setLayout(new MigLayout("", "[grow][grow]", "[]"));
        setBorder(BorderFactory.createTitledBorder("Response list"));

        final DefaultMutableTreeNode top = new DefaultMutableTreeNode("Response list");
        DefaultTreeModel treeModel = new DefaultTreeModel(top);

        responseTree = new JTree(treeModel) {
            @Override
            public String convertValueToText(Object value, boolean selected, boolean expanded, boolean leaf, int row,
                    boolean hasFocus) {
                final DefaultMutableTreeNode node = (DefaultMutableTreeNode)value;
                if (node.getUserObject() instanceof TdpMessage) {
                    final TdpMessage message = (TdpMessage)node.getUserObject();
                    return super.convertValueToText(message.getCommand().getName(), selected, expanded, leaf, row,
                            hasFocus);
                } else {
                    return super.convertValueToText(value, selected, expanded, leaf, row, hasFocus);
                }
            }
        };
        responseTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);

        LOG.debug(responseTree.getFont().getFamily().toString());
        LOG.debug(responseTree.getFont().getFontName().toString());
        LOG.debug("" + responseTree.getFont().getStyle());
        LOG.debug("" + responseTree.getFont().getSize());
        final JScrollPane responseView = new JScrollPane(responseTree);
        add(responseView, "grow");

        final DefaultMutableTreeNode rootNotifications = new DefaultMutableTreeNode("Notifications");
        DefaultTreeModel treeNotificationModel = new DefaultTreeModel(rootNotifications);

        notificationTree = new JTree(treeNotificationModel) {
            @Override
            public String convertValueToText(Object value, boolean selected, boolean expanded, boolean leaf, int row,
                    boolean hasFocus) {
                final DefaultMutableTreeNode node = (DefaultMutableTreeNode)value;
                if (node.getUserObject() instanceof TdpCommunicationError) {
                    final TdpCommunicationError error = (TdpCommunicationError)node.getUserObject();
                    return super.convertValueToText(error.getCommunicationError().getDescription(), selected, expanded,
                            leaf, row, hasFocus);
                } else if (node.getUserObject() instanceof TdpNotification) {
                    final TdpNotification notification = (TdpNotification)node.getUserObject();
                    return super.convertValueToText(notification.getTag().getMessage(), selected, expanded, leaf, row,
                            hasFocus);
                } else {
                    return super.convertValueToText(value, selected, expanded, leaf, row, hasFocus);
                }
            }
        };
        notificationTree.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);

        final JScrollPane notificationView = new JScrollPane(notificationTree);
        add(notificationView, "grow, wrap");

        add(new JLabel("Message data:"), "wrap");
        messageData = new JTextArea(10, 30);
        messageData.setFont(new Font("Tahoma", 0, 11));
        final JScrollPane scrollPane = new JScrollPane(messageData);
        messageData.setEditable(false);
        add(scrollPane, "span 2, wrap, grow");
    }

    @Override
    public final void responseListChanged(List<TdpMessage> responseList) {
        LOG.debug("Build response tree");
        final DefaultMutableTreeNode top = (DefaultMutableTreeNode)responseTree.getModel().getRoot();

        top.removeAllChildren();

        for (TdpMessage response : responseList) {
            final DefaultMutableTreeNode node = new DefaultMutableTreeNode(response);
            top.add(node);

            for (Integer address : response.getAddresses()) {
                node.add(new DefaultMutableTreeNode(address));
            }
        }

        JTreeUtils.expandAll(responseTree, true);

        responseTree.updateUI();
        messageData.setText("");

        return;
    }

    @Override
    public final void clearResponses() {
        final DefaultMutableTreeNode root = (DefaultMutableTreeNode)responseTree.getModel().getRoot();
        root.removeAllChildren();
        responseTree.updateUI();

        final DefaultMutableTreeNode rootNotification = (DefaultMutableTreeNode)notificationTree.getModel().getRoot();
        rootNotification.removeAllChildren();
        notificationTree.updateUI();

        messageData.setText("");
    }

    @Override
    public final void notificationReceived(TdpNotifications notifications) {
        LOG.debug("Build notification tree");
        final DefaultMutableTreeNode top = (DefaultMutableTreeNode)notificationTree.getModel().getRoot();

        top.removeAllChildren();

        for (TdpNotification notification : notifications.getNotifications()) {
            final DefaultMutableTreeNode node = new DefaultMutableTreeNode(notification);
            top.add(node);

            if (notification instanceof TdpCommunicationError) {
                final TdpCommunicationError error = (TdpCommunicationError)notification;
                node.add(new DefaultMutableTreeNode(error));
            }
        }

        JTreeUtils.expandAll(notificationTree, true);

        notificationTree.updateUI();

        return;
    }

    @Override
    public void notificationsChanged(int displayAddress, List<TdpNotification> notificationList) {
        // ignore
    }

    /**
     * @return the commandTree
     */
    public final JTree getResponseTree() {
        return responseTree;
    }

    /**
     * @return the messageData
     */
    public final JTextArea getMessageData() {
        return messageData;
    }
}
