/*
Copyright (c) 2013-2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.communication;

import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.nio.ByteBuffer;
import java.util.Locale;

import org.disperanto.enums.EIpConnectionType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * TCP server to receive messages and send responses.
 * 
 * The server maintains a connection with only one client. When another client connects the
 * server disconnects the previous client.
 */
public class TcpReceiver implements IStartableConnection {

    private static final Logger LOG = LoggerFactory.getLogger(TcpReceiver.class);

    private volatile boolean running;
    private ServerSocket serverSocket;
    private final int port;
    private final int readTimeoutMs;
    private Socket clientSocket;

    /**
     * Creates a new TCP server socket to receive messages.
     * @param port TCP port number.
     */
    public TcpReceiver(int port, int readTimeoutMs) {
        this.port = port;
        this.readTimeoutMs = readTimeoutMs;
    }

    @Override
    public void send(ByteBuffer data) throws IOException {
        data.rewind();
        if (clientSocket != null && !clientSocket.isClosed()) {
            OutputStream out = clientSocket.getOutputStream();
            out.write(data.array());
            out.flush();
        }
    }

    @Override
    public InputStream getInputStream() throws IOException {
        if (!running) {
            return null;
        }
        
        if (clientSocket == null || clientSocket.isClosed()) {
            // Wait for new connection
            clientSocket = serverSocket.accept();
            clientSocket.setSoTimeout(readTimeoutMs);
        }
        return clientSocket.getInputStream();
    }
    
    public String getRemoteIpAddress() {
        if (clientSocket != null && !clientSocket.isClosed()) {
            return clientSocket.getInetAddress().getHostAddress();
        }
        return "";
    }
    
    @Override
    public EIpConnectionType getConnectionType() {
        return EIpConnectionType.TCP;
    }

    @Override
    public String getIpAddress() {
        if (serverSocket != null && !serverSocket.isClosed()) {
            return serverSocket.getInetAddress().getHostAddress();
        }
        return "";
    }

    @Override
    public int getPort() {
        return port;
    }

    @Override
    public void start() throws IOException {
        if (running) {
            return;
        }
        
        LOG.info("Starting TcpReceiver");
        try {
            serverSocket = new ServerSocket(port, 1);
            running = true;
            LOG.info("TcpReceiver started on port " + port);
        } catch (IOException e) {
            LOG.error("Error on opening ServerSocket", e);
            // Throw e anyway, because higher level code may depend on it (e.g. to limit the retry rate).
            throw e;
        }
    }

    @Override
    public void stop() {
        if (running) {
            LOG.info("Stopping TcpReceiver");

            close(clientSocket, "client");
            clientSocket = null;

            close(serverSocket, "server");
            serverSocket = null;

            running = false;
            LOG.info("TcpReceiver stopped");
        } else {
            LOG.info("TcpReceiver not running, so not stopped");
        }
    }

    private void close(Closeable socket, String kind) {
        if (socket != null) {
            try {
                socket.close();
                LOG.info("Closed socket: {}", kind);
            } catch (IOException e) {
                LOG.error(String.format("Error while closing socket: %s", kind), e);
            }
        }
    }

    @Override
    public String toString() {
        return String.format(Locale.ROOT, "%s [port=%s, running=%b]", getClass().getSimpleName(), port, running);
    }
}
