/*
Copyright (c) 2013-2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.communication;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.BindException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.nio.ByteBuffer;

import org.disperanto.enums.EIpConnectionType;
import org.disperanto.utils.ByteBufferUtils;
import org.disperanto.utils.TdpUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Sets up a UDP connection to receive commands and send responses.
 * 
 * The receiver can maintain only one connection to a sender. 
 */
public class UdpReceiver implements IStartableConnection {

    private static final Logger LOGGER = LoggerFactory.getLogger(UdpReceiver.class);

    // return address for UDP message
    private InetAddress ipAddress;
    // return port for UDP message
    private int port;

    private final int timeoutMs;
    private final int receivePort;

    private final Object socketLock = new Object();
    private DatagramSocket socket;

    /**
     * Creates a new UDP receiver.
     *
     * @param receivePort UDP port to listen to.
     * @param timeoutMs Timeout for UDP communication in milliseconds
     */
    public UdpReceiver(int receivePort, int timeoutMs) {
        this.timeoutMs = timeoutMs;
        this.receivePort = receivePort;
    }

    /**
     * Starts the UDP connection.
     */
    @Override
    public final void start() throws IOException {
        try {
            synchronized (socketLock) {
                socket = new DatagramSocket(receivePort);
                socket.setSoTimeout(timeoutMs);
            }
        } catch (BindException be) {
            // Log the port that was used in case of an error
            LOGGER.warn("Unable to bind datagramsocket on port: " + receivePort);
            throw be; // Allow calling code to handle the error
        }
    }

    /**
     * Stops the UDP connection.
     */
    @Override
    public final void stop() {
        synchronized (socketLock) {
            if (socket != null) {
                socket.close();
                socket = null;
            }
        }
    }

    @Override
    public final void send(ByteBuffer data) throws IOException {
        try {
            data.rewind();
            final DatagramPacket packet = new DatagramPacket(data.array(), data.capacity(), ipAddress, port);

            synchronized (socketLock) {
                if (socket == null) {
                    throw new IOException("Socket is closed");
                }
                socket.send(packet);
            }

            LOGGER.trace("Send packet: " + packet.getLength() + " bytes to " + packet.getAddress() + ":" +
                    packet.getPort() + ": " + ByteBufferUtils.toByteString(data));
        } catch (IllegalArgumentException ex) {
            LOGGER.trace("IllegalArgumentException: " + ex.getMessage());
            throw new IOException("Unresolved address: " + ipAddress);
        }
    }

    /**
     * Receives a UDP packet from the same address a packet is sent to. <br>
     * Packets from other IP addresses are skipped (no handling at all).
     *
     * @return UDP packet
     * @throws IOException
     */
    private DatagramPacket receivePacket() throws IOException {
        final ByteBuffer packetData = ByteBuffer.allocate(TdpUtils.MAX_BUFFER);

        final DatagramPacket packet = new DatagramPacket(packetData.array(), packetData.capacity());

        if (socket == null) {
            throw new IOException("No socket");
        }
        socket.receive(packet);

        return packet;
    }
    
    @Override
    public EIpConnectionType getConnectionType() {
        return EIpConnectionType.UDP;
    }

    @Override
    public final String getIpAddress() {
        return ipAddress.getHostAddress();
    }

    @Override
    public int getPort() {
        return port;
    }

    public int getReceivePort() {
        return receivePort;
    }

    public final int getTimeoutMs() {
        return timeoutMs;
    }

    @Override
    public InputStream getInputStream() throws IOException {
        ByteBuffer buffer = receive();
        if (buffer != null && buffer.hasArray()) {
            return new ByteArrayInputStream(buffer.array());
        }
        return null;
    }

    private ByteBuffer receive() throws IOException {
        final DatagramPacket packet = receivePacket();

        // Save return address
        ipAddress = packet.getAddress();
        port = packet.getPort();

        final ByteBuffer messageData = ByteBuffer.allocate(packet.getLength());
        messageData.put(packet.getData(), 0, packet.getLength());

        LOGGER.trace("Received packet: " + packet.getLength() + " bytes: " + ByteBufferUtils.toByteString(messageData));
        return messageData;
    }
}
