/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;

import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpDisplayProperties;
import org.disperanto.enums.ETdpDisplayType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.CompactTlvReader;
import org.disperanto.utils.CompactTlvWriter;


/**
 * Formatter for the DisplayProperties response.
 */
public final class DisplayPropertiesResponse {
    private static final String MESSAGENAME = ETdpCommand.DISPLAY_PROPERTIES.getName();
    private static final int INITIAL_CAPACITY = 100;

    /**
     * Constructor
     */
    private DisplayPropertiesResponse() {
        // Empty constructor
    }

    /**
     * Creates the TLV data buffer for the DisplayProperties response.
     * 
     * @param displayProperties Data of this response message
     * @return Data buffer for the display properties data
     */
    public static ByteBuffer createData(TdpDisplayProperties displayProperties) {
        final TdpFormatter formatter = new TdpFormatter();
        final CompactTlvWriter writer = new CompactTlvWriter(INITIAL_CAPACITY);

        writer.add((byte)ETdpDisplayProperties.PROTOCOL_VERSION.getValue(),
                formatter.setOneByteData(displayProperties.getProtocolVersion(),
                ETdpDisplayProperties.PROTOCOL_VERSION.getDescription(), MESSAGENAME));

        writer.add((byte)ETdpDisplayProperties.DISPLAY_TYPE.getValue(), formatter.setOneByteData(
                displayProperties.getDisplayType().getValue(), ETdpDisplayProperties.DISPLAY_TYPE.getDescription(),
                MESSAGENAME));

        writer.add((byte)ETdpDisplayProperties.SUPPLIER_AND_PRODUCTTYPE.getValue(), formatter.setStringData(
                displayProperties.getSupplierAndProductType(), 40,
                ETdpDisplayProperties.SUPPLIER_AND_PRODUCTTYPE.getDescription(), MESSAGENAME));

        writer.add((byte)ETdpDisplayProperties.SERIALNUMBER.getValue(),
                formatter.setStringData(displayProperties.getSerialNumber(), 20,
                ETdpDisplayProperties.SERIALNUMBER.getDescription(), MESSAGENAME));

        writer.add((byte)ETdpDisplayProperties.SOFTWARE_VERSION.getValue(), formatter.setStringData(
                displayProperties.getSoftwareVersion(), 20, ETdpDisplayProperties.SOFTWARE_VERSION.getDescription(),
                MESSAGENAME));

        if (displayProperties.hasExternalLighting()) {
            writer.add((byte)ETdpDisplayProperties.EXTERNAL_LIGHTING.getValue(), null);
        }

        if (displayProperties.getDisplayType() == ETdpDisplayType.MATRIX) {
            writer.add((byte)ETdpDisplayProperties.DISPLAY_HEIGHT.getValue(), formatter.setMaxTwoBytesData(
                    displayProperties.getDisplayHeightInPixels(),
                    ETdpDisplayProperties.DISPLAY_HEIGHT.getDescription(), MESSAGENAME));

            writer.add((byte)ETdpDisplayProperties.DISPLAY_WIDTH.getValue(), formatter.setMaxTwoBytesData(
                    displayProperties.getDisplayWidthInPixels(), ETdpDisplayProperties.DISPLAY_WIDTH.getDescription(),
                    MESSAGENAME));
        }

        if (displayProperties.getNrFixedImages() != 0) {
            writer.add((byte)ETdpDisplayProperties.NR_FIXED_IMAGES.getValue(), formatter.setMaxTwoBytesData(
                    displayProperties.getNrFixedImages(), ETdpDisplayProperties.NR_FIXED_IMAGES.getDescription(),
                    MESSAGENAME));
        }

        if (displayProperties.getNrWritableImages() != 0) {
            writer.add((byte)ETdpDisplayProperties.NR_WRITABLE_IMAGES.getValue(), formatter.setMaxTwoBytesData(
                    displayProperties.getNrWritableImages(), ETdpDisplayProperties.NR_WRITABLE_IMAGES.getDescription(),
                    MESSAGENAME));
        }

        if (displayProperties.hasSlideShow()) {
            writer.add((byte)ETdpDisplayProperties.SLIDE_SHOW.getValue(),
                    formatter.setOneByteData(displayProperties.getNrSlideShowImages(),
                    ETdpDisplayProperties.SLIDE_SHOW.getDescription(), MESSAGENAME));

        }

        if (displayProperties.getDisplayType() == ETdpDisplayType.MATRIX) {
            if ((displayProperties.getRgbColors() == null) && (displayProperties.getFixedColorPalette() == null)) {
                throw new IllegalArgumentException(
                        "Display properties must contain data for RGB_COLORS or FIXED_COLOR_PALETTE when "
                                + "display type = MATRIXE, neither is present");
            }
            if ((displayProperties.getRgbColors() != null) && (displayProperties.getFixedColorPalette() != null)) {
                throw new IllegalArgumentException(
                        "Display properties must contain data for RGB_COLORS or FIXED_COLOR_PALETTE when "
                                + "display type = MATRIX, both are present");
            }
        }

        if (displayProperties.getRgbColors() != null) {
            writer.add((byte)ETdpDisplayProperties.RGB_COLORS.getValue(),
                    formatter.setByteBufferData(displayProperties.getRgbColors(), 3,
                    ETdpDisplayProperties.RGB_COLORS.getDescription(), MESSAGENAME));
        }
        if (displayProperties.getFixedColorPalette() != null) {
            writer.add((byte)ETdpDisplayProperties.FIXED_COLOR_PALETTE.getValue(),
                    setFixedColorPaletteData(displayProperties.getFixedColorPalette()));
        }

        if (displayProperties.isPngSupported()) {
            writer.add((byte)ETdpDisplayProperties.PNG_SUPPORTED.getValue(), null);
        }

        return writer.getData();
    }

    /**
     * Parses the received data to a TdpDisplayProperties object.<br>
     *
     * Sequence of tags in data is mandatory.
     *
     * @param message Generic message that is received from a display
     * @return Traffic Display properties
     */
    public static TdpDisplayProperties parse(TdpMessage message) throws TdpProtocolException {
        final TdpDisplayProperties properties = new TdpDisplayProperties();
        final TdpFormatter formatter = new TdpFormatter();
        ByteBuffer tagData;

        formatter.checkResponse(message, ETdpCommand.DISPLAY_PROPERTIES);

        // Parse TLV data
        final ByteBuffer messageData = formatter.getMessageData(message);
        final CompactTlvReader tlvReader = new CompactTlvReader(messageData);
        tlvReader.next();

        tagData = getTagData(tlvReader, ETdpDisplayProperties.PROTOCOL_VERSION.getValue(),
                ETdpDisplayProperties.PROTOCOL_VERSION.getDescription(), MESSAGENAME, true);
        properties.setProtocolVersion(formatter.parseOneByteData(tagData,
                ETdpDisplayProperties.PROTOCOL_VERSION.getDescription(), MESSAGENAME));
        tlvReader.next();

        tagData = getTagData(tlvReader, ETdpDisplayProperties.DISPLAY_TYPE.getValue(),
                ETdpDisplayProperties.DISPLAY_TYPE.getDescription(), MESSAGENAME, true);
        properties.setDisplayType(parseDisplayTypeData(tagData, ETdpDisplayProperties.DISPLAY_TYPE.getDescription(),
                MESSAGENAME));
        tlvReader.next();

        tagData = getTagData(tlvReader, ETdpDisplayProperties.SUPPLIER_AND_PRODUCTTYPE.getValue(),
                ETdpDisplayProperties.SUPPLIER_AND_PRODUCTTYPE.getDescription(), MESSAGENAME, true);
        properties.setSupplierAndProductType(formatter.parseStringData(tagData, 40,
                ETdpDisplayProperties.SUPPLIER_AND_PRODUCTTYPE.getDescription(), MESSAGENAME));
        tlvReader.next();

        tagData = getTagData(tlvReader, ETdpDisplayProperties.SERIALNUMBER.getValue(),
                ETdpDisplayProperties.SERIALNUMBER.getDescription(), MESSAGENAME, true);
        properties.setSerialNumber(formatter.parseStringData(tagData, 20,
                ETdpDisplayProperties.SERIALNUMBER.getDescription(), MESSAGENAME));
        tlvReader.next();

        tagData = getTagData(tlvReader, ETdpDisplayProperties.SOFTWARE_VERSION.getValue(),
                ETdpDisplayProperties.SOFTWARE_VERSION.getDescription(), MESSAGENAME, true);
        properties.setSoftwareVersion(formatter.parseStringData(tagData, 20,
                ETdpDisplayProperties.SOFTWARE_VERSION.getDescription(), MESSAGENAME));
        tlvReader.next();

        if (isTagPresent(tlvReader, ETdpDisplayProperties.EXTERNAL_LIGHTING.getValue(),
                ETdpDisplayProperties.EXTERNAL_LIGHTING.getDescription(), MESSAGENAME, false)) {
            properties.setExternalLighting(true);
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpDisplayProperties.DISPLAY_HEIGHT.getValue(),
                ETdpDisplayProperties.DISPLAY_HEIGHT.getDescription(), MESSAGENAME, false);
        if (tagData != null) {
            properties.setDisplayHeightInPixels(formatter.parseMaxTwoBytesData(tagData,
                    ETdpDisplayProperties.DISPLAY_HEIGHT.getDescription(), MESSAGENAME));
            tlvReader.next();
        } else {
            if (properties.getDisplayType() == ETdpDisplayType.MATRIX) {
                throw new TdpProtocolException(
                        "Display properties must contain data for display height when display type = MATRIX");
            }
        }

        tagData = getTagData(tlvReader, ETdpDisplayProperties.DISPLAY_WIDTH.getValue(),
                ETdpDisplayProperties.DISPLAY_WIDTH.getDescription(), MESSAGENAME, false);
        if (tagData != null) {
            properties.setDisplayWidthInPixels(formatter.parseMaxTwoBytesData(tagData,
                    ETdpDisplayProperties.DISPLAY_WIDTH.getDescription(), MESSAGENAME));
            tlvReader.next();
        } else {
            if (properties.getDisplayType() == ETdpDisplayType.MATRIX) {
                throw new TdpProtocolException(
                        "Display properties must contain data for display width when display type = MATRIX");
            }
        }

        tagData = getTagData(tlvReader, ETdpDisplayProperties.NR_FIXED_IMAGES.getValue(),
                ETdpDisplayProperties.NR_FIXED_IMAGES.getDescription(), MESSAGENAME, false);
        if (tagData != null) {
            properties.setNrFixedImages(formatter.parseMaxTwoBytesData(tagData,
                    ETdpDisplayProperties.NR_FIXED_IMAGES.getDescription(), MESSAGENAME));
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpDisplayProperties.NR_WRITABLE_IMAGES.getValue(),
                ETdpDisplayProperties.NR_WRITABLE_IMAGES.getDescription(), MESSAGENAME, false);
        if (tagData != null) {
            properties.setNrWritableImages(formatter.parseMaxTwoBytesData(tagData,
                    ETdpDisplayProperties.NR_WRITABLE_IMAGES.getDescription(), MESSAGENAME));
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpDisplayProperties.SLIDE_SHOW.getValue(),
                ETdpDisplayProperties.SLIDE_SHOW.getDescription(), MESSAGENAME, false);
        if (tagData != null) {
            properties.setSlideShow(true);
            properties.setNrSlideShowImages(formatter.parseOneByteData(tagData,
                    ETdpDisplayProperties.SLIDE_SHOW.getDescription(), MESSAGENAME));
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpDisplayProperties.RGB_COLORS.getValue(),
                ETdpDisplayProperties.RGB_COLORS.getDescription(), MESSAGENAME, false);
        if (tagData != null) {
            properties.setRgbColors(formatter.parseByteBufferData(tagData, 3,
                    ETdpDisplayProperties.RGB_COLORS.getDescription(), MESSAGENAME));
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpDisplayProperties.FIXED_COLOR_PALETTE.getValue(),
                ETdpDisplayProperties.FIXED_COLOR_PALETTE.getDescription(), MESSAGENAME, false);
        if (tagData != null) {
            properties.setFixedColorPalette(parseFixedColorPaletteData(tagData));
            tlvReader.next();
        }

        if (properties.getDisplayType() == ETdpDisplayType.MATRIX) {
            if ((properties.getRgbColors() == null) && (properties.getFixedColorPalette() == null)) {
                throw new TdpProtocolException(
                        "Display properties must contain data for RGB_COLORS or FIXED_COLOR_PALETTE when "
                                + "display type = MATRIX, neither is present");
            }
            if ((properties.getRgbColors() != null) && (properties.getFixedColorPalette() != null)) {
                throw new TdpProtocolException(
                        "Display properties must contain data for RGB_COLORS or FIXED_COLOR_PALETTE when "
                                + "display type = MATRIX, both are present");
            }
        }

        if (isTagPresent(tlvReader, ETdpDisplayProperties.PNG_SUPPORTED.getValue(),
                ETdpDisplayProperties.PNG_SUPPORTED.getDescription(), MESSAGENAME, false)) {
            properties.setPngSupported(true);
        }

        formatter.checkRemainingData(messageData);

        return properties;
    }

    /**
     * Checks the tag that is read and returns the data for this tag.
     *
     * @param tlvReader Reader utility for reading TLV-data
     * @param expectedTagId Value of the tag that is expected to be read
     * @param expectedTagDescription Description of the tag that is expected
     * @param commandName Name of the command of the message that is parsed
     * @param mandatory Flag that indicates if tag is mandatory or not
     */
    protected static ByteBuffer getTagData(final CompactTlvReader tlvReader, int expectedTagId,
            String expectedTagDescription, String commandName, boolean mandatory) throws TdpProtocolException {
        if (isTagPresent(tlvReader, expectedTagId, expectedTagDescription, commandName, mandatory)) {
            return tlvReader.getData();
        } else {
            return null;
        }
    }

    /**
     * Checks if the given tag is present.
     *
     * @param tlvReader Reader utility for reading TLV-data
     * @param expectedTagId Value of the tag that is expected to be read
     * @param expectedTagDescription Description of the tag that is expected
     * @param commandName Name of the command of the message that is parsed
     * @param mandatory Flag that indicates if tag is mandatory or not
     */
    protected static boolean isTagPresent(final CompactTlvReader tlvReader, int expectedTagId,
            String expectedTagDescription, String commandName, boolean mandatory) throws TdpProtocolException {
        final int id = tlvReader.getId();

        final ETdpDisplayProperties tag = ETdpDisplayProperties.convertValue(id);
        if (tag == null) {
            throw new TdpProtocolException("Unknown tag in data of " + commandName + " message");
        }
        if (tag.getValue() != expectedTagId) {
            if (mandatory) {
                throw new TdpProtocolException("Mandatory tag " + expectedTagDescription + " missing in data of " +
                        commandName + " message");
            } else {
                // tag is optional and not present
                return false;
            }
        }

        return true;
    }

    /**
     * Parses the data for the display type
     *
     * @param data Byte buffer containing the data to be parsed
     * @return
     * @throws TdpProtocolException
     */
    private static ETdpDisplayType parseDisplayTypeData(ByteBuffer data, String tagDescription,
            String messageDescription) throws TdpProtocolException {
        final TdpFormatter formatter = new TdpFormatter();

        final ETdpDisplayType displayType = ETdpDisplayType.convertValue(formatter.parseOneByteData(data,
                tagDescription, messageDescription));

        if (displayType != null) {
            return displayType;
        } else {
            throw new TdpProtocolException("Unknown value for DisplayType in DisplayProperties response");
        }
    }

    /**
     * Parses the data for the fixed color palette.
     *
     * @param data Byte buffer containing the data to be parsed
     */
    private static ByteBuffer parseFixedColorPaletteData(ByteBuffer data) throws TdpProtocolException {
        // expected data length is of multiple of 3 bytes
        if ((data.capacity() % 3) != 0) {
            throw new TdpProtocolException(
                    "Wrong data length received for FixedColorPalette tag from DisplayProperties response "
                            + "(must be multiple of 3)");
        }

        return data;
    }

    /**
     * Sets the data for the fixed color palette in a byte buffer
     *
     * @param fixedColorPalette Fixed color palette used by display
     * @throws TdpProtocolException
     */
    private static ByteBuffer setFixedColorPaletteData(ByteBuffer fixedColorPalette) {
        // expected data length is 3 bytes
        if (fixedColorPalette.capacity() != 3) {
            throw new IllegalArgumentException(
                    "Wrong data length for FixedColorPalette tag of DisplayProperties command");
        }

        return fixedColorPalette;
    }
}
