/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import org.disperanto.data.IImageOperation;
import org.disperanto.data.TdpClearRectangle;
import org.disperanto.data.TdpCopyImage;
import org.disperanto.data.TdpInitializeWorkingMemory;
import org.disperanto.data.TdpLoadImage;
import org.disperanto.data.TdpManipulateMemorySlot;
import org.disperanto.data.TdpStoreImage;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpImageType;
import org.disperanto.enums.ETdpManipulateMemorySlot;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.ByteBufferUtils;
import org.disperanto.utils.CompactTlvReader;
import org.disperanto.utils.CompactTlvWriter;
import org.disperanto.utils.VlqUtils;


/**
 * Formatter for the ManipulateMemorySlot command.
 */
public final class ManipulateMemorySlotCommand {
    private static final String MESSAGENAME = ETdpCommand.MANIPULATE_MEMORY_SLOT.getName();
    private static final int INITIAL_CAPACITY = 2000;

    /**
     * Constructor
     */
    private ManipulateMemorySlotCommand() {
        // Empty constructor
    }

    /**
     * Creates the TLV data buffer for the ManipulateMemorySlot command.
     *
     * @param manipulateMemorySlot Data of this command message
     * @return Data buffer for the display properties data
     */
    public static ByteBuffer createData(TdpManipulateMemorySlot manipulateMemorySlot)  {
        final CompactTlvWriter writer = new CompactTlvWriter(INITIAL_CAPACITY);

        for (IImageOperation operation : manipulateMemorySlot.getOperations()) {
            switch (operation.getType()) {
            case INITIALIZE_WORKING_MEMORY :
                writer.add((byte)ETdpManipulateMemorySlot.INITIALIZE_WORKING_MEMORY.getValue(),
                        getInitializeWorkingMemoryData((TdpInitializeWorkingMemory)operation));
                break;
            case CLEAR_RECTANGLE :
                writer.add((byte)ETdpManipulateMemorySlot.CLEAR_RECTANGLE.getValue(),
                        getClearRectangleData((TdpClearRectangle)operation));
                break;
            case COPY_IMAGE :
                writer.add((byte)ETdpManipulateMemorySlot.COPY_IMAGE.getValue(),
                        getCopyImageData((TdpCopyImage)operation));
                break;
            case LOAD_IMAGE :
                writer.add((byte)ETdpManipulateMemorySlot.LOAD_IMAGE.getValue(),
                        getLoadImageData((TdpLoadImage)operation));
                break;
            case STORE_IMAGE :
                writer.add((byte)ETdpManipulateMemorySlot.STORE_IMAGE.getValue(),
                        getStoreImageData((TdpStoreImage)operation));
                break;
            default :
                throw new IllegalStateException("Unknown operationType: " + operation.getType());
            }
        }

        return writer.getData();
    }

    /**
     * Parses the received data to a TdpManipulateMemorySlot object.<br>
     * <br>
     * Sequence of tags in data is mandatory.
     *
     * @param message Generic message that is received from a display
     * @return Traffic Display properties
     */
    public static TdpManipulateMemorySlot parse(TdpMessage message) throws TdpProtocolException {
        final TdpFormatter formatter = new TdpFormatter();
        final TdpManipulateMemorySlot manipulateSlot = new TdpManipulateMemorySlot();
        ByteBuffer tagData;

        formatter.checkCommand(message, ETdpCommand.MANIPULATE_MEMORY_SLOT);

        final ByteBuffer messageData = formatter.getMessageData(message);
        final CompactTlvReader tlvReader = new CompactTlvReader(messageData);
        tlvReader.next();

        List<IImageOperation> operations = new ArrayList<IImageOperation>();

        tagData = getTagData(tlvReader, ETdpManipulateMemorySlot.INITIALIZE_WORKING_MEMORY.getValue(),
                ETdpManipulateMemorySlot.INITIALIZE_WORKING_MEMORY.getDescription(), MESSAGENAME, true);
        operations.add(parseInitializeWorkingMemoryData(tagData));

        while (tlvReader.next()) {
            int id = tlvReader.getId();
            ETdpManipulateMemorySlot tag = ETdpManipulateMemorySlot.convertValue(id);
            tagData = tlvReader.getData();
            operations.add(parseData(tag, tagData));
        }

        manipulateSlot.setOperations(operations);

        formatter.checkRemainingData(messageData);

        return manipulateSlot;
    }

    private static IImageOperation parseData(ETdpManipulateMemorySlot tag, ByteBuffer tagData)
            throws TdpProtocolException {
        switch (tag) {
        case INITIALIZE_WORKING_MEMORY :
            return parseInitializeWorkingMemoryData(tagData);
        case CLEAR_RECTANGLE :
            return parseClearRectangleData(tagData);
        case COPY_IMAGE :
            return parseCopyImageData(tagData);
        case LOAD_IMAGE :
            return parseLoadImageData(tagData);
        case STORE_IMAGE :
            return parseStoreImageData(tagData);
        default :
            throw new IllegalStateException("Unknown tag: " + tag);
        }
    }

    // TDOD LBA: The functions getTagData and isTagPresent are used in each message with TLV-data. This functions should
    // be coded once in the TdpFormatter, but in that case I have an issue with convertValue() in isTagPresent. How to
    // solve this?
    /**
     * Checks the tag that is read and returns the data for this tag.
     *
     * @param tlvReader Reader utility for reading TLV-data
     * @param expectedTagId Value of the tag that is expected to be read
     * @param expectedTagDescription Description of the tag that is expected
     * @param commandName Name of the command of the message that is parsed
     * @param mandatory Flag that indicates if tag is mandatory or not
     */
    protected static ByteBuffer getTagData(final CompactTlvReader tlvReader, int expectedTagId,
            String expectedTagDescription, String commandName, boolean mandatory) throws TdpProtocolException {
        if (isTagPresent(tlvReader, expectedTagId, expectedTagDescription, commandName, mandatory)) {
            return tlvReader.getData();
        } else {
            return null;
        }
    }

    /**
     * Checks if the given tag is present.
     *
     * @param tlvReader Reader utility for reading TLV-data
     * @param expectedTagId Value of the tag that is expected to be read
     * @param expectedTagDescription Description of the tag that is expected
     * @param commandName Name of the command of the message that is parsed
     * @param mandatory Flag that indicates if tag is mandatory or not
     */
    protected static boolean isTagPresent(final CompactTlvReader tlvReader, int expectedTagId,
            String expectedTagDescription, String commandName, boolean mandatory) throws TdpProtocolException {
        final int id = tlvReader.getId();

        final ETdpManipulateMemorySlot tag = ETdpManipulateMemorySlot.convertValue(id);
        if (tag == null) {
            throw new TdpProtocolException("Unknown tag in data of " + commandName + " message");
        }
        if (tag.getValue() != expectedTagId) {
            if (mandatory) {
                throw new TdpProtocolException("Mandatory tag " + expectedTagDescription + " missing in data of " +
                        commandName + " message");
            } else {
                // tag is optional and not present
                return false;
            }
        }

        return true;
    }

    /**
     * Parses the data for the InitializeWorkingMemory command
     *
     * @param data Data to be set
     * @return
     * @throws TdpProtocolException
     */
    private static TdpInitializeWorkingMemory parseInitializeWorkingMemoryData(ByteBuffer data)
            throws TdpProtocolException {

        final TdpInitializeWorkingMemory initializeMemory = new TdpInitializeWorkingMemory();

        initializeMemory.setWidth(VlqUtils.parseVlqData(data));
        initializeMemory.setHeight(VlqUtils.parseVlqData(data));

        return initializeMemory;

    }

    /**
     * Parses the data for the ClearRectangle command
     *
     * @param data Data to be set
     * @return
     * @throws TdpProtocolException
     */
    private static TdpClearRectangle parseClearRectangleData(ByteBuffer data) throws TdpProtocolException {

        final TdpClearRectangle clearRectangle = new TdpClearRectangle();

        clearRectangle.setLeft(VlqUtils.parseVlqData(data));
        clearRectangle.setTop(VlqUtils.parseVlqData(data));
        clearRectangle.setWidth(VlqUtils.parseVlqData(data));
        clearRectangle.setHeight(VlqUtils.parseVlqData(data));

        return clearRectangle;
    }

    /**
     * Parses the data for the LoadImage command
     *
     * @param data Data to be set
     * @return
     * @throws TdpProtocolException
     */
    private static TdpLoadImage parseLoadImageData(ByteBuffer data) throws TdpProtocolException {

        final TdpLoadImage loadImage = new TdpLoadImage();

        loadImage.setLeft(VlqUtils.parseVlqData(data));
        loadImage.setTop(VlqUtils.parseVlqData(data));

        loadImage.setImageType(parseImageTypeData(data));
        final byte[] remainingData = new byte[data.capacity() - data.position()];
        data.get(remainingData);
        loadImage.setImage(remainingData);

        return loadImage;
    }

    /**
     * Parses the data for the CopyImage command
     *
     * @param data Data to be set
     * @return
     * @throws TdpProtocolException
     */
    private static TdpCopyImage parseCopyImageData(ByteBuffer data) throws TdpProtocolException {

        final TdpCopyImage copyImage = new TdpCopyImage();

        copyImage.setLeft(VlqUtils.parseVlqData(data));
        copyImage.setTop(VlqUtils.parseVlqData(data));
        copyImage.setMemorySlot(VlqUtils.parseVlqData(data));

        return copyImage;
    }

    /**
     * Parses the data for the StoreImage command
     *
     * @param data Data to be set
     * @return
     * @throws TdpProtocolException
     */
    private static TdpStoreImage parseStoreImageData(ByteBuffer data) throws TdpProtocolException {

        final TdpStoreImage storeImage = new TdpStoreImage();

        storeImage.setMemorySlot(VlqUtils.parseVlqData(data));

        return storeImage;
    }

    /**
     * Parses the data for the display type
     *
     * @param data Byte buffer containing the data to be parsed
     * @return
     * @throws TdpProtocolException
     */
    private static ETdpImageType parseImageTypeData(ByteBuffer data) throws TdpProtocolException {
        final ETdpImageType imageType = ETdpImageType.convertValue(data.get());

        if (imageType != null) {
            return imageType;
        } else {
            throw new TdpProtocolException(
                    "Unknown value for ImageType in LoadImage of ManipulateWorkingMemory command");
        }
    }

    /**
     * Sets the data for the InitializeWorkingMemory command
     *
     * @param initializeMemory Data to be set
     * @return
     * @throws TdpProtocolException
     */
    private static ByteBuffer getInitializeWorkingMemoryData(TdpInitializeWorkingMemory initializeMemory) {

        final ByteBuffer data = ByteBuffer.allocate(VlqUtils.MAX_VLQ_BYTES * 2);

        data.put(VlqUtils.getVlqBytes(initializeMemory.getWidth()));
        data.put(VlqUtils.getVlqBytes(initializeMemory.getHeight()));

        return ByteBufferUtils.truncateAtCurrentPosition(data);
    }

    /**
     * Sets the data for the ClearRectangle command
     *
     * @param clearRectangle Data to be set
     * @return
     * @throws TdpProtocolException
     */
    private static ByteBuffer getClearRectangleData(TdpClearRectangle clearRectangle) {

        final ByteBuffer data = ByteBuffer.allocate(VlqUtils.MAX_VLQ_BYTES * 4);

        data.put(VlqUtils.getVlqBytes(clearRectangle.getLeft()));
        data.put(VlqUtils.getVlqBytes(clearRectangle.getTop()));
        data.put(VlqUtils.getVlqBytes(clearRectangle.getWidth()));
        data.put(VlqUtils.getVlqBytes(clearRectangle.getHeight()));

        return ByteBufferUtils.truncateAtCurrentPosition(data);
    }

    /**
     * Sets the data for the LoadImage command
     *
     * @param loadImage Data to be set
     * @return
     * @throws TdpProtocolException
     */
    private static ByteBuffer getLoadImageData(TdpLoadImage loadImage) {
        final TdpFormatter formatter = new TdpFormatter();
        byte[] image = loadImage.getImage();
        final ByteBuffer data = ByteBuffer.allocate((VlqUtils.MAX_VLQ_BYTES * 2) + 1 + image.length);

        data.put(VlqUtils.getVlqBytes(loadImage.getLeft()));
        data.put(VlqUtils.getVlqBytes(loadImage.getTop()));
        data.put(formatter.setOneByteData(loadImage.getImageType().getValue(),
                loadImage.getImageType().getDescription(), MESSAGENAME));
        data.put(image);

        return ByteBufferUtils.truncateAtCurrentPosition(data);
    }

    /**
     * Sets the data for the CopyImage command
     *
     * @param copyImage Data to be set
     * @return
     * @throws TdpProtocolException
     */
    private static ByteBuffer getCopyImageData(TdpCopyImage copyImage) {

        final ByteBuffer data = ByteBuffer.allocate(VlqUtils.MAX_VLQ_BYTES * 3);

        data.put(VlqUtils.getVlqBytes(copyImage.getLeft()));
        data.put(VlqUtils.getVlqBytes(copyImage.getTop()));
        data.put(VlqUtils.getVlqBytes(copyImage.getMemorySlot()));

        return ByteBufferUtils.truncateAtCurrentPosition(data);
    }

    /**
     * Sets the data for the StoreImage command
     *
     * @param storeImage Data to be set
     * @return
     * @throws TdpProtocolException
     */
    private static ByteBuffer getStoreImageData(TdpStoreImage storeImage) {

        final ByteBuffer data = ByteBuffer.allocate(VlqUtils.MAX_VLQ_BYTES);

        data.put(VlqUtils.getVlqBytes(storeImage.getMemorySlot()));

        return ByteBufferUtils.truncateAtCurrentPosition(data);
    }
}
