/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import org.disperanto.data.TdpCommunicationError;
import org.disperanto.data.TdpNotification;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpCommunicationError;
import org.disperanto.enums.ETdpNotification;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.CompactTlvReader;
import org.disperanto.utils.CompactTlvWriter;


/**
 * Represents a received notification from a display.
 */
public final class NotificationResponse {

    /**
     * Constructor
     */
    private NotificationResponse() {
        // Empty constructor
    }

    /**
     * Creates the data buffer for the Notification response.
     *
     * @param notifications Data of this response message
     * @return Data buffer for the timeout settings
     */
    public static ByteBuffer createData(List<TdpNotification> notifications) {
        final TdpFormatter formatter = new TdpFormatter();
        final CompactTlvWriter writer = new CompactTlvWriter(notifications.size() * 2);

        for (TdpNotification notification : notifications) {
            if (notification.getTag() == ETdpNotification.COMMUNICATION_ERROR) {
                final TdpCommunicationError communicationErrorNotification = (TdpCommunicationError)notification;
                if (communicationErrorNotification.getCommunicationError() == null) {
                    throw new IllegalArgumentException("No data for CommunicationError notification");
                }

                writer.add((byte)notification.getTag().getValue(), formatter.setOneByteData(
                        communicationErrorNotification.getCommunicationError().getValue(),
                        notification.getTag().getMessage(), "Notification"));
            } else {
                writer.add((byte)notification.getTag().getValue(), null);
            }
        }

        return writer.getData();
    }

    /**
     * Parses the received Disperanto message to a TdpNotification object
     *
     * @param message Generic message that is received from a display
     * @return Traffic Display notification
     */
    public static List<TdpNotification> parse(TdpMessage message) throws TdpProtocolException {
        final TdpFormatter formatter = new TdpFormatter();
        final List<TdpNotification> notifications = new ArrayList<TdpNotification>();

        formatter.checkResponse(message, ETdpCommand.NOTIFICATION);

        // Parse data containing notifications
        final ByteBuffer messageData = formatter.getMessageData(message);
        final CompactTlvReader tlvReader = new CompactTlvReader(messageData);
        while (tlvReader.next()) {
            final int id = tlvReader.getId();
            final ByteBuffer data = tlvReader.getData();

            // set notification
            final ETdpNotification tag = ETdpNotification.convertValue(id);
            if (tag == null) {
                // unknown notification received
                throw new TdpProtocolException("Unexpected tag in data of Notification response");
            }

            if (tag == ETdpNotification.COMMUNICATION_ERROR) {
                // parse data for communication error
                notifications.add(parseCommunicationError(data));
            } else {
                // all notifications without data
                notifications.add(new TdpNotification(tag));
            }
        }

        return notifications;
    }

    /**
     * Parses the data of an communication error.
     *
     * @param data Data of a communication error
     * @param notifications Notifications object in which the data of the communication error is stored
     * @throws TdpProtocolException
     */
    private static TdpCommunicationError parseCommunicationError(ByteBuffer data) throws TdpProtocolException {
        final TdpCommunicationError notification = new TdpCommunicationError();
        // expected data length 1
        if (data.capacity() != 1) {
            throw new TdpProtocolException(
                    "Wrong data length received for CommunicationError from Notification response");
        }

        data.rewind();
        byte value = data.get();
        final ETdpCommunicationError error = ETdpCommunicationError.convertValue(value);
        if (error == null) {
            throw new TdpProtocolException("Unexpected value (" + Integer.toHexString(value) 
                    + ") in data of communication error of Notification response");
        } else {
            notification.setCommunicationError(error);
        }

        return notification;
    }
}
