/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.io.ByteArrayOutputStream;
import java.nio.ByteBuffer;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;

import org.disperanto.data.TdpSetText;
import org.disperanto.data.TdpSetText.TdpTextLine;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpSetText;
import org.disperanto.enums.ETdpTextAlignment;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.VlqUtils;

/**
 * Formatter for the SetText command.
 */
public final class SetTextCommand {
    public static final int MAX_CHARACTERS_ROW = 255;
    public static final int MAX_ROWS = 255;

    private static final String MESSAGENAME = ETdpCommand.SET_TEXT.getName();
    private static final int INITIAL_CAPACITY = 50;
    private static final CharsetEncoder ASCII_ENCODER = StandardCharsets.US_ASCII.newEncoder();

    /**
     * Constructor
     */
    private SetTextCommand() {
        // Empty constructor
    }

    /**
     * Creates the TLV data buffer for the SetText command.
     *
     * @param setText Data of this command message
     * @return Data buffer for the set text data
     */
    public static ByteBuffer createData(TdpSetText setText)  {
        if (setText.getLines() == null) {
            throw new IllegalArgumentException("Lines is mandatory in SetText command");
        }
        if (setText.getLines().size() > MAX_ROWS) {
            throw new IllegalArgumentException("Maximum of 255 Lines allowed in SetText command");
        }

        final TdpFormatter formatter = new TdpFormatter();
        final ByteArrayOutputStream tdpOut = new ByteArrayOutputStream(INITIAL_CAPACITY);

        tdpOut.write(setText.getLines().size());
        for (TdpTextLine line : setText.getLines()) {
            // write alignment
            tdpOut.write((byte)ETdpSetText.ALIGNMENT.getValue());
            byte[] alignmentData = formatter.setOneByteData(line.getAlignment().getValue(), 
                    ETdpSetText.ALIGNMENT.getDescription(), MESSAGENAME).array();
            tdpOut.write(alignmentData, 0, alignmentData.length);
            // write text
            tdpOut.write((byte)ETdpSetText.TEXT.getValue());
            byte[] lineData = setTextData(line);
            tdpOut.write(lineData, 0, lineData.length);
        }
        return ByteBuffer.wrap(tdpOut.toByteArray());
    }

    /**
     * Sets the data for the text.<br>
     *
     * @param data Byte buffer containing the data to be parsed
     * @return Communication timeout mode
     * @throws TdpProtocolException
     */
    private static byte[] setTextData(TdpTextLine line) {
        if (line.getText() == null) {
            throw new IllegalArgumentException("Text is mandatory in SetText command");
        }

        final TdpFormatter formatter = new TdpFormatter();
        final ByteArrayOutputStream tdpOut = new ByteArrayOutputStream(INITIAL_CAPACITY);

        if (!ASCII_ENCODER.canEncode(line.getText())) {
            throw new IllegalArgumentException("Text line contains non-ASCII characters");
        }

        byte[] nrOfChars = VlqUtils.getVlqBytes(line.getText().length()).array();
        tdpOut.write(nrOfChars, 0, nrOfChars.length);
        byte[] text = formatter
                .setStringData(line.getText(), MAX_CHARACTERS_ROW, ETdpSetText.TEXT.getDescription(), MESSAGENAME)
                .array();
        tdpOut.write(text, 0, text.length);
        return tdpOut.toByteArray();
    }

    /**
     * Parses the received data to a TdpSetText object.<br>
     * <br>
     * Sequence of tags in data is mandatory.
     *
     * @param message Generic message that is received from a display
     * @return TdpSetText
     */
    public static TdpSetText parse(TdpMessage message) throws TdpProtocolException {
        final TdpFormatter formatter = new TdpFormatter();
        final TdpSetText setTextCommand = new TdpSetText();

        formatter.checkCommand(message, ETdpCommand.SET_TEXT);

        final ByteBuffer messageData = formatter.getMessageData(message);
        int nrOfLines = messageData.get();

        List<TdpTextLine> lines = new ArrayList<>();

        for (int i = 0; i < nrOfLines; i++) {
            // parse alignment
            ETdpSetText tag = ETdpSetText.convertValue(messageData.get());
            if (ETdpSetText.ALIGNMENT != tag) {
                throw new TdpProtocolException("Mandatory Alignment tag expected in data of SetText message");
            }
            ETdpTextAlignment alignment = parseTextAlignmentData(messageData);
            // parse text line
            tag = ETdpSetText.convertValue(messageData.get());
            if (ETdpSetText.TEXT != tag) {
                throw new TdpProtocolException("Mandatory Text tag expected in data of SetText message");
            }
            int length = VlqUtils.parseVlqData(messageData);
            if (length > MAX_CHARACTERS_ROW) {
                throw new TdpProtocolException("Length of Text line too long of SetText message");
            }
            byte[] textData = new byte[length];
            messageData.get(textData, 0, length);
            if (!ASCII_ENCODER.canEncode(new String(textData, StandardCharsets.UTF_8))) {
                throw new IllegalArgumentException("Text line contains non-ASCII characters");
            }
            lines.add(new TdpTextLine(alignment, new String(textData)));
        }
        setTextCommand.setLines(lines);

        formatter.checkRemainingData(messageData);

        return setTextCommand;
    }

    /**
     * Parse the data for the alignment
     *
     * @param data Byte buffer containing the data to be parsed
     * @return
     * @throws TdpProtocolException
     */
    private static ETdpTextAlignment parseTextAlignmentData(ByteBuffer data) throws TdpProtocolException {
        final ETdpTextAlignment alignment = ETdpTextAlignment.convertValue(data.get());

        if (alignment != null) {
            return alignment;
        } else {
            throw new TdpProtocolException("Unknown value for Alignment in TdpTextLine of SetText command");
        }
    }
}
