/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.disperanto.data.TdpStatus;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpStatus;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.ByteBufferUtils;
import org.disperanto.utils.CompactTlvReader;
import org.disperanto.utils.CompactTlvWriter;
import org.disperanto.utils.VlqUtils;


/**
 * Formatter for the Status response.
 */
public final class StatusResponse {
    private static final String MESSAGENAME = ETdpCommand.STATUS.getName();
    private static final int INITIAL_CAPACITY = 100;

    /**
     * Constructor
     */
    private StatusResponse() {
        // Empty constructor
    }

    /**
     * Creates the TLV data buffer for the Status response.
     *
     * @param status Data of this response message
     * @return Data buffer for the status data
     */
    public static ByteBuffer createData(TdpStatus status) {
        final TdpFormatter formatter = new TdpFormatter();
        final CompactTlvWriter writer = new CompactTlvWriter(INITIAL_CAPACITY);

        writer.add((byte)ETdpStatus.SHOWN_IMAGES.getValue(), setShownImagesData(status.getShownImages()));

        if (status.getBrightnessInPercentage() != null) {
            writer.add((byte)ETdpStatus.BRIGHTNESS.getValue(), formatter.setPercentageData(
                    status.getBrightnessInPercentage(), ETdpStatus.BRIGHTNESS.getDescription(), MESSAGENAME));
        }

        if (status.getIntensityExternalLighting() != null) {
            writer.add((byte)ETdpStatus.INTENSITY_EXTERNAL_LIGHTING.getValue(), formatter.setPercentageData(
                    status.getIntensityExternalLighting(), ETdpStatus.INTENSITY_EXTERNAL_LIGHTING.getDescription(),
                    MESSAGENAME));
        }

        if (status.getMeasuredLightIntensity().size() > 0) {
            writer.add((byte)ETdpStatus.MEASURED_LIGHT_INTENSITY.getValue(),
                    setMeasuredLightIntensityData(status.getMeasuredLightIntensity()));
        }

        if (status.getGpsPosition() != null) {
            writer.add((byte)ETdpStatus.GPS_POSITION.getValue(), formatter.setStringData(status.getGpsPosition(), 40,
                    ETdpStatus.GPS_POSITION.getDescription(), MESSAGENAME));
        }

        if (status.getTemperature() != null) {
            writer.add((byte)ETdpStatus.TEMPERATURE.getValue(), formatter.setOneByteData(status.getTemperature(),
                    ETdpStatus.TEMPERATURE.getDescription(), MESSAGENAME));
        }

        if (status.getHeating() != null) {
            writer.add((byte)ETdpStatus.HEATING.getValue(),
                    setHeatingData(status.getHeating(), ETdpStatus.HEATING.getDescription(), MESSAGENAME));
        }

        if (status.getCooling() != null) {
            writer.add((byte)ETdpStatus.COOLING.getValue(),
                    setCoolingData(status.getCooling(), ETdpStatus.COOLING.getDescription(), MESSAGENAME));
        }

        return writer.getData();
    }

    /**
     * Parses the received data to a TdpStatus object.
     *
     * @param message Generic message that is received from a display
     * @return Traffic Display status
     */
    public static TdpStatus parse(TdpMessage message) throws TdpProtocolException {
        final TdpStatus tdpStatus = new TdpStatus();
        final TdpFormatter formatter = new TdpFormatter();
        ByteBuffer tagData;

        formatter.checkResponse(message, ETdpCommand.STATUS);

        // Parse tlv data
        final ByteBuffer messageData = formatter.getMessageData(message);
        final CompactTlvReader tlvReader = new CompactTlvReader(messageData);
        tlvReader.next();

        tagData = getTagData(tlvReader, ETdpStatus.SHOWN_IMAGES.getValue(), ETdpStatus.SHOWN_IMAGES.getDescription(),
                MESSAGENAME, true);
        tdpStatus.setShownImages(parseShownImagesData(tagData));
        tlvReader.next();

        tagData = getTagData(tlvReader, ETdpStatus.BRIGHTNESS.getValue(), ETdpStatus.BRIGHTNESS.getDescription(),
                MESSAGENAME, false);
        if (tagData != null) {
            tdpStatus.setBrightnessInPercentage(formatter.parsePercentageData(tagData,
                    ETdpStatus.BRIGHTNESS.getDescription(), MESSAGENAME));
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpStatus.INTENSITY_EXTERNAL_LIGHTING.getValue(),
                ETdpStatus.INTENSITY_EXTERNAL_LIGHTING.getDescription(), MESSAGENAME, false);
        if (tagData != null) {
            tdpStatus.setIntensityExternalLighting(formatter.parsePercentageData(tagData,
                    ETdpStatus.INTENSITY_EXTERNAL_LIGHTING.getDescription(), MESSAGENAME));
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpStatus.MEASURED_LIGHT_INTENSITY.getValue(),
                ETdpStatus.MEASURED_LIGHT_INTENSITY.getDescription(), MESSAGENAME, false);
        if (tagData != null) {
            tdpStatus.setMeasuredLightIntensity(parseMeasuredLightIntensityData(tagData));
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpStatus.GPS_POSITION.getValue(), ETdpStatus.GPS_POSITION.getDescription(),
                MESSAGENAME, false);
        if (tagData != null) {
            tdpStatus.setGpsPosition(formatter.parseStringData(tagData, tagData.capacity(),
                    ETdpStatus.GPS_POSITION.getDescription(), MESSAGENAME));
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpStatus.TEMPERATURE.getValue(), ETdpStatus.TEMPERATURE.getDescription(),
                MESSAGENAME, false);
        if (tagData != null) {
            tdpStatus.setTemperature(formatter.parseOneByteDataSigned(tagData, ETdpStatus.TEMPERATURE.getDescription(),
                    MESSAGENAME));
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpStatus.HEATING.getValue(), ETdpStatus.HEATING.getDescription(),
                MESSAGENAME, false);
        if (tagData != null) {
            tdpStatus.setHeating(parseHeatingData(tagData, ETdpStatus.HEATING.getDescription(), MESSAGENAME));
            tlvReader.next();
        }

        tagData = getTagData(tlvReader, ETdpStatus.COOLING.getValue(), ETdpStatus.COOLING.getDescription(),
                MESSAGENAME, false);
        if (tagData != null) {
            tdpStatus.setCooling(parseCoolingData(tagData, ETdpStatus.COOLING.getDescription(), MESSAGENAME));
        }

        formatter.checkRemainingData(messageData);

        return tdpStatus;
    }

    /**
     * Checks the tag that is read and returns the data for this tag.
     *
     * @param tlvReader Reader utility for reading TLV-data
     * @param expectedTagId Value of the tag that is expected to be read
     * @param expectedTagDescription Description of the tag that is expected
     * @param commandName Name of the command of the message that is parsed
     * @param mandatory Flag that indicates if tag is mandatory or not
     */
    protected static ByteBuffer getTagData(final CompactTlvReader tlvReader, int expectedTagId,
            String expectedTagDescription, String commandName, boolean mandatory) throws TdpProtocolException {
        if (isTagPresent(tlvReader, expectedTagId, expectedTagDescription, commandName, mandatory)) {
            return tlvReader.getData();
        } else {
            return null;
        }
    }

    /**
     * Checks if the given tag is present.
     *
     * @param tlvReader Reader utility for reading TLV-data
     * @param expectedTagId Value of the tag that is expected to be read
     * @param expectedTagDescription Description of the tag that is expected
     * @param commandName Name of the command of the message that is parsed
     * @param mandatory Flag that indicates if tag is mandatory or not
     */
    protected static boolean isTagPresent(final CompactTlvReader tlvReader, int expectedTagId,
            String expectedTagDescription, String commandName, boolean mandatory) throws TdpProtocolException {
        final int id = tlvReader.getId();

        final ETdpStatus tag = ETdpStatus.convertValue(id);
        if (tag == null) {
            throw new TdpProtocolException("Unknown tag in data of " + commandName + " message");
        }
        if (tag.getValue() != expectedTagId) {
            if (mandatory) {
                throw new TdpProtocolException("Mandatory tag " + expectedTagDescription + " missing in data of " +
                        commandName + " message");
            } else {
                // tag is optional and not present
                return false;
            }
        }

        return true;
    }

    /**
     * Parses the data for the shown image(s) on the display.
     *
     * @param data Byte buffer containing the data to be parsed
     */
    private static Map<Integer, Short> parseShownImagesData(ByteBuffer data) throws TdpProtocolException {
        final Map<Integer, Short> map = new HashMap<Integer, Short>();

        data.rewind();
        while (data.hasRemaining()) {
            // read slot from VLQ byte(s)
            final int slot = VlqUtils.parseVlqData(data);
            final short crc = data.getShort();

            map.put(slot, crc);
        }

        return map;
    }

    /**
     * Sets the data for the shown image(s) in a byte buffer
     *
     * @param shownImages Map of the slots of the shown image(s). Key = slot, value = CRC of image in slot
     */
    private static ByteBuffer setShownImagesData(Map<Integer, Short> shownImages) {
        final ByteBuffer shownImagesData = ByteBuffer.allocate(shownImages.size() * (VlqUtils.MAX_VLQ_BYTES + 2));

        for (Entry<Integer, Short> entry : shownImages.entrySet()) {
            final Integer slot = entry.getKey();
            final Short crc = entry.getValue();

            shownImagesData.put(VlqUtils.getVlqBytes(slot.intValue()));
            shownImagesData.putShort(crc.shortValue());
        }

        return ByteBufferUtils.truncateAtCurrentPosition(shownImagesData);
    }

    /**
     * Parses the data for the shown image(s) on the display.
     *
     * @param data Byte buffer containing the data to be parsed
     */
    private static List<Integer> parseMeasuredLightIntensityData(ByteBuffer data) throws TdpProtocolException {
        final List<Integer> list = new ArrayList<Integer>();

        data.rewind();
        while (data.hasRemaining()) {
            final int value = data.get() & 0xFF;

            if ((value < 0) || (value > 100)) {
                throw new TdpProtocolException("Unexpected value for MeasuredLightIntensity data of Status response (" +
                        value + "). Expected is percentage (0..100).");
            }

            list.add(value);
        }

        return list;
    }

    /**
     * Sets the data for the shown image(s) in a byte buffer.
     *
     * @param shownImages Map of the slots of the shown image(s). Key = slot, value = CRC of image in slot
     */
    private static ByteBuffer setMeasuredLightIntensityData(List<Integer> measuredData) {
        final ByteBuffer shownImagesData = ByteBuffer.allocate(measuredData.size());

        for (Integer i : measuredData) {
            final byte value = i.byteValue();

            if ((value < 0) || (value > 100)) {
                throw new IllegalArgumentException(
                        "Unexpected value for measured light intensity data of Status response (" + value +
                                "). Expected is percentage (0..100).");
            }

            shownImagesData.put(value);
        }

        return ByteBufferUtils.truncateAtCurrentPosition(shownImagesData);
    }

    /**
     * parses the data for the heating in a byte buffer.
     * 
     * @param data Byte buffer containing the data to be parsed
     * @param tagDescription Description of the tag of the TLV data that is read
     * @param messageDescription Description of the message
     * @return Value that is parsed
     */
    private static Integer parseHeatingData(ByteBuffer data, String tagDescription, String messageDescription)
            throws TdpProtocolException {
        final TdpFormatter formatter = new TdpFormatter();
        final int value = formatter.parseOneByteData(data, tagDescription, messageDescription);

        if ((value != 0) && (value != 1)) {
            throw new TdpProtocolException("Unexpected value for Heating data of Status response (" + value +
                    "). Expected is 0 or 1.");
        }

        return Integer.valueOf(value);
    }

    /**
     * Sets the data for the heating in a byte buffer.
     *
     * @param value value that indicates if heating is off or on
     */
    private static ByteBuffer setHeatingData(int value, String tagDescription, String messageDescription) {
        final TdpFormatter formatter = new TdpFormatter();

        if ((value != 0) && (value != 1)) {
            throw new IllegalArgumentException("Unexpected value for Heating data of Status response (" + value +
                    "). Expected is 0 or 1.");
        }

        return formatter.setOneByteData(value, tagDescription, messageDescription);
    }

    /**
     * parses the data for the cooling in a byte buffer.
     *
     * @param data Byte buffer containing the data to be parsed
     * @param tagDescription Description of the tag of the tlv data that is read
     * @param messageDescription Description of the message
     * @return Value that is parsed
     */
    private static Integer parseCoolingData(ByteBuffer data, String tagDescription, String messageDescription)
            throws TdpProtocolException {
        final TdpFormatter formatter = new TdpFormatter();

        final int value = formatter.parseOneByteData(data, tagDescription, messageDescription);

        if ((value != 0) && (value != 1)) {
            throw new TdpProtocolException("Unexpected value for Cooling data of Status response (" + value +
                    "). Expected is 0 or 1.");
        }

        return Integer.valueOf(value);
    }

    /**
     * Sets the data for the cooling in a byte buffer.
     *
     * @param value value that indicates if heating is off or on
     */
    private static ByteBuffer setCoolingData(int value, String tagDescription, String messageDescription) {
        final TdpFormatter formatter = new TdpFormatter();

        if ((value != 0) && (value != 1)) {
            throw new IllegalArgumentException("Unexpected value for Cooling data of Status response (" + value +
                    "). Expected is 0 or 1.");
        }

        return formatter.setOneByteData(value, tagDescription, messageDescription);
    }
}
