/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.ByteBufferUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This represents a Disperanto message.<br>
 *
 * It holds the raw data. Use the TdpProtocol class to decode this data.
 */
public class TdpMessage {
    private static final Logger LOGGER = LoggerFactory.getLogger(TdpMessage.class);

    private ETdpMessageType messageType;
    private ETdpCommand commandType;
    private List<Integer> addresses = new ArrayList<Integer>();
    private ByteBuffer data;
    private int messageNumber;

    /**
     * Constructor: Creates a TDP message.
     *
     * @param messageType Message type of the TDP message
     * @param commandType The identification of the message
     * @param addresses the display addresses that the message is sent to/received from
     * @param data The message data
     */
    public TdpMessage(ETdpMessageType messageType, ETdpCommand commandType, List<Integer> addresses, ByteBuffer data) {
        this.messageType = messageType;
        this.commandType = commandType;
        this.addresses = addresses;
        this.data = data;
    }

    /**
     * Constructor: Creates an empty message.
     */
    public TdpMessage() {
        // Empty constructor
    }

    /**
     * Gets the message type.
     *
     * @return the messageType
     */
    public final ETdpMessageType getMessageType() {
        return messageType;
    }

    /**
     * @param messageType the messageType to set
     */
    public final void setMessageType(ETdpMessageType messageType) {
        this.messageType = messageType;
    }

    /**
     * Gets the command.
     *
     * @return the command
     */
    public final ETdpCommand getCommand() {
        return commandType;
    }

    /**
     * @param command the command to set
     */
    public final void setCommand(ETdpCommand command) {
        this.commandType = command;
    }

    /**
     * Gets the list of display addresses.
     *
     * @return the address list
     */
    public final List<Integer> getAddresses() {
        return addresses;
    }

    /**
     * @param addresses the addresses to set
     */
    public final void setAddresses(List<Integer> addresses) {
        this.addresses = addresses;
    }

    /**
     * @param address the address to add to the list of addresses
     */
    public final void addAddress(Integer address) {
        addresses.add(address);
    }

    /**
     * Gets the (read only) data.
     *
     * @return the data of the message
     */
    public final ByteBuffer getData() {
        return data;
    }

    /**
     * @param data the data to set
     */
    public final void setData(ByteBuffer data) {
        this.data = data;
    }

    /**
     * Gets the message number.
     *
     * @return the messageNumber
     */
    public final int getMessageNumber() {
        return messageNumber;
    }

    /**
     * Sets the message number
     *
     * @param messageNumber the message number to set
     */
    public final void setMessageNumber(int messageNumber) {
        this.messageNumber = messageNumber;
    }

    @Override
    public final String toString() {
        final StringBuilder builder = new StringBuilder(messageType.toString()).append(" ");
        builder.append(commandType.getName());
        builder.append("\n\tAddresses ").append(addresses.toString());

        if (data != null && data.hasRemaining()) {
            data.rewind();
            builder.append("\n\tdata: 0x");
            builder.append(ByteBufferUtils.toByteString(data));
        } else {
            builder.append("\n\tNo data");
        }

        return builder.toString();
    }

    public final String toReadableString() {
        final StringBuilder builder = new StringBuilder(messageType.toString()).append(" ");
        builder.append(commandType.getName());
        builder.append("\n\tAddresses ").append(addresses.toString());
        builder.append("\n\tMessage number ").append(messageNumber);

        if (data != null) {
            builder.append("\n\t").append(getReadableMessageData());
        } else {
            builder.append("\n\tNo data");
        }

        return builder.toString();
    }

    private String getReadableMessageData() {
        try {
            if (commandType == ETdpCommand.DISPLAY_PROPERTIES && messageType == ETdpMessageType.RESPONSE) {
                return DisplayPropertiesResponse.parse(this).toString();
            }
            if (commandType == ETdpCommand.STATUS && messageType == ETdpMessageType.RESPONSE) {
                return StatusResponse.parse(this).toString();
            }
            if (commandType == ETdpCommand.SET_COMMUNICATION_TIMEOUT && messageType == ETdpMessageType.COMMAND) {
                return SetTimeoutCommand.parse(this).toString();
            }
            if (commandType == ETdpCommand.SET_BRIGHTNESS && messageType == ETdpMessageType.COMMAND) {
                return SetBrightnessCommand.parse(this).toString();
            }
            if (commandType == ETdpCommand.SET_EXTERNAL_LIGHTING && messageType == ETdpMessageType.COMMAND) {
                return SetExternalLightingCommand.parse(this).toString();
            }
            if (commandType == ETdpCommand.DIAGNOSTICS && messageType == ETdpMessageType.RESPONSE) {
                return DiagnosticsResponse.parse(this).toString();
            }
            if (commandType == ETdpCommand.MANIPULATE_MEMORY_SLOT && messageType == ETdpMessageType.COMMAND) {
                return ManipulateMemorySlotCommand.parse(this).toString();
            }
            if (commandType == ETdpCommand.MANIPULATE_MEMORY_SLOT && messageType == ETdpMessageType.RESPONSE) {
                return "" + ManipulateMemorySlotResponse.parse(this);
            }
            if (commandType == ETdpCommand.CALCULATE_CRC_IMAGES && messageType == ETdpMessageType.COMMAND) {
                return CalculateCrcImagesCommand.parse(this).toString();
            }
            if (commandType == ETdpCommand.CALCULATE_CRC_IMAGES && messageType == ETdpMessageType.RESPONSE) {
                return CalculateCrcImagesResponse.parse(this).toString();
            }
            if (commandType == ETdpCommand.SHOW_IMAGE && messageType == ETdpMessageType.COMMAND) {
                return ShowImageCommand.parse(this).toString();
            }
            if (commandType == ETdpCommand.SHOW_IMAGE && messageType == ETdpMessageType.RESPONSE) {
                return "" + ShowImageResponse.parse(this);
            }
            if (commandType == ETdpCommand.START_SLIDE_SHOW && messageType == ETdpMessageType.COMMAND) {
                return StartSlideShowCommand.parse(this).toString();
            }
            if (commandType == ETdpCommand.START_SLIDE_SHOW && messageType == ETdpMessageType.RESPONSE) {
                return StartSlideShowResponse.parse(this).toString();
            }
            if (commandType == ETdpCommand.SET_TEXT && messageType == ETdpMessageType.COMMAND) {
                return SetTextCommand.parse(this).toString();
            }
        } catch (TdpProtocolException e) {
            LOGGER.warn("TdpProtocolException: '" + e.getMessage() + "'while parsing data for " +
                    messageType.getName() + " " + commandType.getName() + " with message number " + messageNumber);
        }

        return null;
    }

}
