/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.protocol;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.TimeUnit;

import org.disperanto.communication.IConnection;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.messages.TdpMessage;
import org.disperanto.utils.TdpUtils;
import org.disperanto.utils.VlqUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Converts Disperanto message to a data packet that can be send.<br>
 *
 * It waits for the responses and creates a list of Disperanto message containing the responses and
 * notifications.<br>
 *
 * All protected functions have the scope 'protected' to support unit testing.
 */
public abstract class TdpProtocol {
    private static final Logger LOGGER = LoggerFactory.getLogger(TdpProtocol.class);

    private static final int VERSION = 1;
    private static final int FIXED_NR_OF_DATA = 5; // start byte, sequence number, command byte and crc (2 bytes)

    // Cyclic number for messages. Runs from 1 to 255.
    private int messageNumber = 0;

    /**
     * Constructor
     */
    public TdpProtocol() {
        // Empty constructor
    }

    /**
     * @return Name of the protocol
     */
    public final String name() {
        return "TrafficDisplayProtocol";
    }

    /**
     * @return Version of the protocol
     */
    public final int version() {
        return VERSION;
    }

    @Override
    public final String toString() {
        return name() + "_v" + version();
    }

    /**
     * Sends a list of Traffic Display messages to a receiver.
     *
     * @param connection Connection used to send the data
     * @param messages list of messages that are sent
     */
    protected final void sendMessages(IConnection connection, List<TdpMessage> messages) throws IOException {
        // Create a buffer for the packet to be send
        final ByteBuffer buffer = TdpUtils.createBuffer(calculatePacketLength(messages));

        final Iterator<TdpMessage> it = messages.iterator();
        while (it.hasNext()) {
            final TdpMessage message = it.next();
            buffer.put(createMessageData(message, !it.hasNext()));
        }

        // Send packet
        connection.send(buffer);

        LOGGER.debug("A list of " + messages.size() + " Disperanto messages is sent");
    }

    /**
     * Receives a list of Traffic Display messages.
     *
     * @param connection Connection used to receive the data
     * @return List of Traffic display messages
     * @throws java.net.SocketTimeoutException If the read operation timed out.
     */
    protected final List<TdpMessage> receiveMessages(IConnection connection) throws IOException, TdpProtocolException {
        InputStream inputStream = connection.getInputStream();
        if (inputStream == null) {
            throw new IOException("Connection closed");
        }

        long startTimeNanos = System.nanoTime();
        List<TdpMessage> received = parseData(inputStream);
        long finishTimeNanos = System.nanoTime();

        LOGGER.debug("A list of " + received.size() + " Disperanto messages is received after " +
                TimeUnit.NANOSECONDS.toSeconds(finishTimeNanos - startTimeNanos) + " seconds");

        return received;
    }

    /**
     * Creates a byte buffer for a Disperanto message.<br>
     * The message number of the message is also set.
     *
     * @param message Disperanto message that has to be converted to byte data in an TDP packet
     * @param lastMessage Flag for indication the last message of a TDP packet
     * @return Byte buffer for a Traffic Display message in an TDP packet
     */
    protected final ByteBuffer createMessageData(TdpMessage message, boolean lastMessage) {

        final ByteBuffer msgBuffer = ByteBuffer.allocate(calculateMessageLength(message));

        // get next sequence in case message is COMMAND (response copies message number from command)
        if (message.getMessageType() == ETdpMessageType.COMMAND) {
            message.setMessageNumber(getMessageNumber());
        }

        msgBuffer.put((byte)createStartByte(message, lastMessage));
        msgBuffer.put((byte)message.getMessageNumber());
        msgBuffer.put(getAddressBuffer(message.getAddresses()));
        msgBuffer.put((byte)message.getCommand().getValue());
        if (message.getData() != null) {
            msgBuffer.put(VlqUtils.getVlqBytes(message.getData().capacity()));
        } else {
            // data length is 0
            msgBuffer.put((byte)0x00);
        }
        if (message.getData() != null) {
            final ByteBuffer messageData = message.getData();
            messageData.rewind();
            msgBuffer.put(messageData);
        }
        msgBuffer.putShort(TdpUtils.calculateCrc(msgBuffer.array()));

        msgBuffer.rewind();
        return msgBuffer;
    }

    /**
     * Parses data read from the {@link InputStream} until the end of the stream into a list of Traffic display
     * messages.<br>
     *
     * @param in InputStream to read from
     * @return List of Traffic Display messages
     */
    protected final List<TdpMessage> parseData(InputStream in) throws TdpProtocolException, IOException {
        boolean lastMessage = false;
        final List<TdpMessage> readMessages = new ArrayList<TdpMessage>();

        while (!lastMessage) {
            final TdpMessage message = new TdpMessage();
            lastMessage = parseMessageData(in, message);
            readMessages.add(message);
        }

        return readMessages;
    }

    /**
     * Parses data read from the {@link InputStream} and reads one Traffic Display message from the data.
     *
     * @param in InputStream to read from.
     * @param message Traffic Display message that is read from data.
     * @return true if this is the last message according the protocol. Otherwise false.
     */
    protected final boolean parseMessageData(InputStream in, TdpMessage message) throws IOException,
            TdpProtocolException {

        ByteBuffer receivedData = ByteBuffer.allocate(TdpUtils.MAX_BUFFER);

        final byte startByte = readOneByte(in, receivedData);
        message.setMessageType(parseMessageType(startByte));
        message.setMessageNumber(readOneByte(in, receivedData));
        for (int i = 0; i < parseNrAddresses(startByte); i++) {
            message.addAddress(Integer.valueOf(readOneByte(in, receivedData)));
        }

        final byte commandByte = readOneByte(in, receivedData);
        message.setCommand(parseCommand(commandByte));
        final int dataLength = VlqUtils.parseVlqData(readVlqData(in, receivedData));
        // Read data
        if (dataLength >= 0) {
            final ByteBuffer data = ByteBuffer.allocate(dataLength);
            for (int i = 0; i < dataLength; i++) {
                data.put(readOneByte(in, receivedData));
            }
            message.setData(data);
        }

        // Read CRC
        ByteBuffer crcReadBuffer = ByteBuffer.allocate(2);
        crcReadBuffer.order(ByteOrder.BIG_ENDIAN);
        crcReadBuffer.put(readOneByte(in, receivedData));
        crcReadBuffer.put(readOneByte(in, receivedData));
        final short crcRead = crcReadBuffer.getShort(0);

        // Check crc
        // Copy byte array for message (inclusive two crc bytes).
        // Calculate crc for byte array and check the value against the read crc.
        receivedData.rewind();
        final int messageLength = calculateMessageLength(parseNrAddresses(startByte), dataLength);
        final ByteBuffer crcBuffer = ByteBuffer.allocate(messageLength);
        receivedData.get(crcBuffer.array(), 0, messageLength);
        final short crcCalculated = TdpUtils.calculateCrc(crcBuffer.array());
        if (crcCalculated != crcRead) {
            throw new TdpProtocolException("CRC of received message is incorrect. Received CRC = " + crcRead +
                    "; Calculated CRC = " + crcCalculated);
        }

        return parseLastMessageFlag(startByte);
    }

    private ByteBuffer readVlqData(InputStream in, ByteBuffer receivedData) throws IOException {
        ByteBuffer buffer = ByteBuffer.allocate(TdpUtils.MAX_BUFFER);
        byte readByte = readOneByte(in, receivedData);
        buffer.put(readByte);

        while (VlqUtils.hasMoreVLQBytes(readByte)) {
            readByte = readOneByte(in, receivedData);
            buffer.put(readByte);
        }
        buffer.rewind();
        return buffer;
    }

    /**
     * Calculates the packet length for the given list of Traffic Display messages.
     *
     * @param messages List of Traffic Display messages.
     * @return Packet length
     */
    private int calculatePacketLength(List<TdpMessage> messages) {
        int length = 0;

        for (TdpMessage message : messages) {
            length += calculateMessageLength(message);
        }

        return length;
    }

    /**
     * Calculates the message length.<br>
     * <br>
     * Message structure:
     * <ul>
     * <li>byte 0: start byte
     * <li>byte 1: sequence number
     * <li>byte 2 .. n+1: list of n display addresses
     * <li>byte n+2: command id
     * <li>byte n+3..m: VLQ-length
     * <li>bytes m+1..p: data
     * <li>bytes p+1..p+2: crc
     * </ul>
     * Fixed number of bytes = 5: start byte, messageNumber (1 byte), commandId (1 byte), CRC (2 bytes)<br>
     * <br>
     * Variable number of addresses (1..n)<br>
     * Optional VLQ-bytes if data length > 2<br>
     * Variable data length (0..x)<br>
     * <br>
     *
     * @param message Traffic Display message.
     * @return Message length
     */
    protected final int calculateMessageLength(TdpMessage message) {
        // VLQ data length field is used

        // Length is put in 7 bits from the VLQ-bytes. The highest bit of every byte indicates if there are more
        // bytes
        if (message.getData() == null) {
            // data length is 0
            return FIXED_NR_OF_DATA + message.getAddresses().size() + VlqUtils.countVlqBytes(0);
        } else {
            return FIXED_NR_OF_DATA + message.getAddresses().size() +
                    VlqUtils.countVlqBytes(message.getData().capacity()) + message.getData().capacity();
        }
    }

    /**
     * Calculates message length based on number of addresses and tlv data length.
     *
     * @param nrAddresses Number of addresses the message is sent to
     * @param dataLength Length of the tlv data in the message
     * @return message length
     */
    private int calculateMessageLength(int nrAddresses, int dataLength) {
        return FIXED_NR_OF_DATA + nrAddresses + VlqUtils.countVlqBytes(dataLength) + dataLength;
    }

    /**
     * Creates the first byte of the message in the packet to send.<br>
     * <br>
     * Byte structure:
     * <ul>
     * <li>bit 7: Message type of TDP message
     * <li>bit 6: Flag for last message of packet
     * <li>bit 0-5: List of addresses (max = 0x1F)
     * </ul>
     *
     * @param message Traffic Display message
     * @param lastMessage flag for last message of the packet
     * @return first byte of the message
     */

    protected final int createStartByte(TdpMessage message, boolean lastMessage) {
        if (message.getAddresses().size() > 0x1F) {
            throw new IllegalArgumentException("Too many addresses in message (" + message.getAddresses().size() +
                    ". Maximum is 31");
        }

        if (!lastMessage) {
            return ((message.getMessageType().getValue() << 7) | message.getAddresses().size());
        } else {
            // include bit for last message (bit 6)
            return ((message.getMessageType().getValue() << 7) | 0x40 | message.getAddresses().size());
        }
    }

    /**
     * Puts the addresses of the displays in a byte buffer
     *
     * @param addresses List of display addresses
     * @return Buffer containing addresses
     */
    protected final ByteBuffer getAddressBuffer(List<Integer> addresses) {
        final ByteBuffer buffer = TdpUtils.createBuffer(addresses.size());

        // loop through addresses and put them in the buffer
        for (Integer address : addresses) {
            if (address.intValue() > 255) {
                throw new IllegalArgumentException("Value of display address does not fit in byte");
            }

            buffer.put(address.byteValue());
        }

        // reset buffer
        buffer.rewind();
        return buffer;
    }

    /**
     * Gets the sequence number of the message.
     *
     * This is a cyclic number that runs from 1 to 255
     */
    private byte getMessageNumber() {
        messageNumber++;

        if (messageNumber > 255) {
            messageNumber = 1;
        }

        return (byte)messageNumber;
    }

    /**
     * Parses the message type from the given byte.
     *
     * @param startByte First byte of the message
     * @return Message type
     */
    protected final ETdpMessageType parseMessageType(byte startByte) throws TdpProtocolException {
        // The message type is set in bit 7
        final ETdpMessageType messageType = ETdpMessageType.convertValue((startByte & 0x80) >> 7);
        if (messageType == null) {
            throw new TdpProtocolException("Unknown message type received (byte = " + startByte + ")");
        }
        return messageType;
    }

    /**
     * Reads the flag for the last message of the received data buffer.
     *
     * @param startByte First byte of the message
     * @return Flag for last message in data
     */
    protected final boolean parseLastMessageFlag(byte startByte) {
        // The flag for last message is set in bit 6
        return ((((startByte & 0x40) >> 6) == 1) ? true : false);
    }

    /**
     * Parses the number of addresses from the given byte
     *
     * @param startByte First byte of the message
     * @return number of addresses in message
     */
    protected final int parseNrAddresses(int startByte) {
        // The number of addresses is set in bit 0-5
        return (startByte & 0x3F);
    }

    /**
     * Parses the command from the given byte.
     *
     * @param commandByte Command byte of the message
     */
    protected final ETdpCommand parseCommand(int commandByte) {
        return ETdpCommand.convertValue(commandByte);
    }

    private byte readOneByte(InputStream in, ByteBuffer readData) throws IOException {
        int readByte = in.read();
        if (readByte != -1) {
            readData.put((byte)readByte);
            return (byte)readByte;
        }

        throw new IOException("End of stream reached without reading a complete message");
    }
}
