/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.protocol;

import java.io.IOException;
import java.net.SocketTimeoutException;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import org.disperanto.communication.IStartableConnection;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.messages.TdpMessage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Receives Disperanto commands.<br>
 */
public class TdpProtocolReceiver extends TdpProtocol implements Runnable {

    private static final Logger LOGGER = LoggerFactory.getLogger(TdpProtocolReceiver.class);

    private final List<ITdpReceiveListener> receiveListeners = new CopyOnWriteArrayList<ITdpReceiveListener>();
    private final IStartableConnection connection;
    
    private AtomicBoolean active = new AtomicBoolean(true);
    private final Semaphore stopped = new Semaphore(0);

    /**
     * Constructor
     * 
     * @param connection connection
     */
    public TdpProtocolReceiver(IStartableConnection connection) {
        this.connection = connection;
    }

    @Override
    public final void run() {
        LOGGER.debug("Receive thread started");
        try {
            while (active.get()) {
                receiveLoop();
            }
        } finally {
            stopped.release();
        }
    }

    private void receiveLoop() {
        try {
            // Read incoming commands
            final List<TdpMessage> commands = receiveMessages(connection);
            if (!commands.isEmpty()) {
                // Notify listeners
                notifyReceive(commands);
            }
        } catch (SocketTimeoutException ste) {
            // Timeout exception received.
            LOGGER.trace("Socket timeout in ({}): {}", connection, ste);
        } catch (TdpProtocolException | IOException e) {
            LOGGER.info("Exception in ({})", connection, e);
            connection.stop();
            if (active.get()) {
                try {
                    connection.start();
                } catch (IOException e1) {
                    // ignore, keep retrying but sleep a while
                    trySleep(500);
                }
            }
        }
    }

    private void trySleep(int millis) {
        try {
            Thread.sleep(millis);
        } catch (InterruptedException ie) {
            // ignore
        }
    }

    /**
     * Notifies the receiver that it should stop, but doesn't wait for the stop to complete
     * @return {@code true} if the receiver was running.
     */
    public boolean cancel() {
        return active.getAndSet(false);
    }

    /**
     * Stops the thread and closes the underlying connection
     */
    public final void stop() throws InterruptedException {
        if (cancel()) {
            // Stop the connection to break out of the blocking receive call.
            connection.stop();

            // join run method
            if (!stopped.tryAcquire(10, TimeUnit.SECONDS)) {
                throw new InterruptedException(getClass().getSimpleName() + " stop timed out");
            }
        }
    }

    /**
     * Adds a listener for received commands.
     * 
     * @param listener Listener for received commands
     */
    public final void addReceiveListener(ITdpReceiveListener listener) {
        receiveListeners.add(listener);
    }

    /**
     * Sends a list of Traffic Display messages.
     * 
     * @param messages List of traffic Display messages that will be send
     */
    public final void send(List<TdpMessage> messages) throws IOException {
        sendMessages(connection, messages);
    }

    /**
     * Notifies a receive to the listeners.
     */
    private void notifyReceive(List<TdpMessage> commands) throws TdpProtocolException, IOException {
        for (ITdpReceiveListener listener : receiveListeners) {
            listener.commandReceived(commands);
        }
    }

}
