/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.utils;

import java.nio.ByteBuffer;
import java.nio.charset.Charset;

/**
 * Generic utilities for Disperanto.
 */
public abstract class ByteBufferUtils {

    private static Charset charset = Charset.forName("UTF-8");

    /**
     * Converts the given ByteBuffer into a hex string.
     * 
     * @param buffer The ByteBuffer to convert
     * @return The hex string
     */
    public static String toByteString(ByteBuffer buffer) {
        if (buffer == null) {
            return null;
        }

        final StringBuilder sb = new StringBuilder();
        final int oldPosition = buffer.position();

        buffer.rewind();
        while (buffer.hasRemaining()) {

            final String hex = Integer.toHexString(0x0100 + (buffer.get() & 0x00FF)).substring(1);
            sb.append((hex.length() < 2 ? "0" : "") + hex + " ");
        }

        buffer.position(oldPosition);
        return sb.toString();
    }

    /**
     * Creates a new ByteBuffer that is a copy of the given buffer, but truncated at the current position.
     * 
     * @param original ByteBuffer that is copied and truncated
     * @return Truncated buffer
     */

    public static ByteBuffer truncateAtCurrentPosition(ByteBuffer original) {
        original.flip();

        final ByteBuffer newBuffer = original.slice();
        newBuffer.flip();

        // complete new buffer is written, so set the limit to the capacity.
        newBuffer.limit(newBuffer.capacity());
        return newBuffer;
    }

    /**
     * Converts a string to a byte buffer.
     * 
     * @param msg String that is converted
     * @return Byte buffer containing the string data
     */
    public static ByteBuffer stringToByteBuffer(String msg) {
        return charset.encode(msg);
    }
    
    /**
     * Converts a byte buffer to a string.
     * 
     * @param buffer Buffer that is converted
     * @return String containing the byte buffer
     */
    public static String byteBufferToString(ByteBuffer buffer) {
        String data = "";

        final int oldPosition = buffer.position();
        data = charset.decode(buffer).toString();

        // reset buffer's position to its original so it is not altered:
        buffer.position(oldPosition);
        return data;
    }
}
