/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.utils;

import java.awt.image.BufferedImage;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * Generic utilities for handling the Disperanto.
 */
public abstract class TdpUtils {

    // TODO #MM-1453: MAX_BUFFER changed for FAT Nijmegen. Largest display is 384 x 256. 
    //                3 bytes per color ==> 294912 bytes
    //                It would be better if this buffer size is dynamically dependent on the display size. 
    public static final int MAX_BUFFER = 300000;

    /**
     * Initial value to use when calculating a CRC.
     */
    private static final int CRC_INITIAL = 0xFFFF;

    /**
     * Creates the buffer for the message bytes.
     *
     * @param size Buffer size
     * @return Buffer
     */
    public static ByteBuffer createBuffer(int size) {
        final ByteBuffer buffer = ByteBuffer.allocate(size);
        // The default order of a byte buffer is BIG_ENDIAN
        buffer.order(ByteOrder.BIG_ENDIAN);
        return buffer;
    }

    /**
     * Calculates CRC16 for images
     * @param image the images to calculate CRC for
     * @return CRC
     */
    public static short calculateCrc(BufferedImage image) {
        int crc = CRC_INITIAL;

        if (image == null) {
            // handle as empty image (0 x 0)
            return (short)crc;
        }

        for (int y = 0; y < image.getHeight(); ++y) {
            for (int x = 0; x < image.getWidth(); ++x) {
                int pixel = image.getRGB(x, y);
                crc = updateCrc(crc, (byte)((pixel >> 16) & 0xFF));
                crc = updateCrc(crc, (byte)((pixel >> 8) & 0xFF));
                crc = updateCrc(crc, (byte)((pixel >> 0) & 0xFF));
            }
        }
        return (short)crc;
    }

    /**
     * Calculates CRC16-CCITT.<br>
     *
     * The given buffer is inclusive two CRC bytes. These CRC bytes are not used in the CRC calculation.
     * 
     * @param data Bytes to calculate CRC on.
     * @return CRC
     */
    public static short calculateCrc(byte[] data) {
        int crc = CRC_INITIAL;

        // The used buffer (data) already has two bytes reserved for the crc.
        // These bytes must not be included in the crc calculation.
        for (int j = 0; j < data.length - 2; j++) {
            crc = updateCrc(crc, data[j]);
        }

        return (short)crc;
    }

    /**
     * CRC calculation part.
     * 
     * @param crc the previously returned CRC, or CRC_INITIAL for the first call
     * @param b the byte to use
     * @return the updated CRC value
     */
    private static int updateCrc(int crc, byte b) {
        final int polynomial = 0x1021; // 0001 0000 0010 0001 (0, 5, 12)

        for (int i = 0; i < 8; i++) {
            final boolean bit = ((b >> (7 - i) & 1) == 1);
            final boolean c15 = ((crc >> 15 & 1) == 1);
            crc <<= 1;
            if (c15 ^ bit) {
                crc ^= polynomial;
            }
        }

        return crc;
    }
}
