/*
Copyright (c) 2013-2019, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.disperanto.data.TdpSetText;
import org.disperanto.data.TdpSetText.TdpTextLine;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.enums.ETdpTextAlignment;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.TdpUtils;
import org.disperanto.utils.VlqUtils;
import org.hamcrest.CoreMatchers;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * JUnit tests for class TdpSetTextCommand.
 */
public class TdpCommandSetTextTest {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    /**
     * Tests the function parseData()
     */
    @Test
    public final void testParseData() {
        final String textLine = "TdpCommandSetTextTest";

        // correct message (only mandatory fields)
        try {
            final ByteBuffer data = TdpUtils.createBuffer(5 + 4 + 2 * textLine.length());
            data.put((byte)2); // 1 text line
            data.put((byte)0x00); // alignment tag
            data.put((byte)0x00); // LEFT
            data.put((byte)0x01); // text tag
            data.put(VlqUtils.getVlqBytes(textLine.length())); // text length
            data.put(textLine.getBytes()); // text
            data.put((byte)0x00); // alignment tag
            data.put((byte)0x02); // RIGHT
            data.put((byte)0x01); // text tag
            data.put(VlqUtils.getVlqBytes(textLine.length())); // text length
            data.put(textLine.getBytes()); // text

            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.COMMAND);
            message.setCommand(ETdpCommand.SET_TEXT);
            message.addAddress(1);
            message.setData(data);

            TdpSetText tdpSetText = SetTextCommand.parse(message);

            Assert.assertEquals(2, tdpSetText.getLines().size());
            Assert.assertEquals(ETdpTextAlignment.LEFT, tdpSetText.getLines().get(0).getAlignment());
            Assert.assertEquals(textLine, tdpSetText.getLines().get(0).getText());
            Assert.assertEquals(ETdpTextAlignment.CENTER, tdpSetText.getLines().get(1).getAlignment());
            Assert.assertEquals(textLine, tdpSetText.getLines().get(1).getText());
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests the function parseData(): missing alignment tag
     */
    @Test
    public final void testParseDataException1() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(
                CoreMatchers.containsString("Mandatory Alignment tag expected in data of SetText message"));

        // incorrect message (first tag id must be alignment)
        final ByteBuffer data = TdpUtils.createBuffer(2);
        data.put((byte)1); // 1 text line
        data.put((byte)0x01); // text tag

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_TEXT);
        message.addAddress(1);
        message.setData(data);

        SetTextCommand.parse(message);
    }

    /**
     * Tests the function parseData(): invalid alignment
     */
    @Test
    public final void testParseDataException2() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(
                CoreMatchers.containsString("Unknown value for Alignment in TdpTextLine of SetText command"));

        // incorrect message (first tag id must be alignment)
        final ByteBuffer data = TdpUtils.createBuffer(3);
        data.put((byte)1); // 1 text line
        data.put((byte)0x00); // alignment tag
        data.put((byte)0x03); // ??

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_TEXT);
        message.addAddress(1);
        message.setData(data);

        SetTextCommand.parse(message);
    }

    /**
     * Tests the function parseData(): invalid tag
     */
    @Test
    public final void testParseDataException3() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(
                CoreMatchers.containsString("Mandatory Text tag expected in data of SetText message"));

        // incorrect message (first tag id must be alignment)
        final ByteBuffer data = TdpUtils.createBuffer(4);
        data.put((byte)1); // 1 text line
        data.put((byte)0x00); // alignment tag
        data.put((byte)0x02); // CENTER
        data.put((byte)0x00); // alignment tag

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_TEXT);
        message.addAddress(1);
        message.setData(data);

        SetTextCommand.parse(message);
    }

    /**
     * Tests the function parseData(): no data
     */
    @Test
    public final void testParseDataException4() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data in message"));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.SET_TEXT);
        message.addAddress(1);
        message.setData(null);

        SetTextCommand.parse(message);
    }

    /**
     * Tests the function CreateData() with maximum values
     */
    @Test
    public final void testCreateDataMaximum() {
        TdpSetText setText = new TdpSetText();

        char[] charArray = new char[255];
        Arrays.fill(charArray, 'A');

        List<TdpTextLine> lines = new ArrayList<>();
        for (int i = 0; i < 255; i++) {
            lines.add(new TdpTextLine(ETdpTextAlignment.CENTER, new String(charArray)));
        }
        setText.setLines(lines);
        SetTextCommand.createData(setText);
    }
}
