/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.util.LinkedHashMap;

import org.disperanto.data.TdpStartSlideShow;
import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.enums.ETdpSlideShowMode;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.TdpUtils;
import org.disperanto.utils.VlqUtils;
import org.hamcrest.CoreMatchers;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * JUnit tests for class StartSlideShowCommand.
 */
public class TdpCommandStartSlideShowTest {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    /**
     * Tests the function parse()
     */
    @Test
    public final void testParse() {
        TdpStartSlideShow settings = new TdpStartSlideShow();
        final ETdpSlideShowMode mode = ETdpSlideShowMode.SHOW_ONCE;
        final int imageSlot1 = 128;
        final Integer showTime1 = 16384;
        final int imageSlot2 = 127;
        final Integer showTime2 = 16383;

        try {
            // correct message
            final ByteBuffer data = TdpUtils.createBuffer(6);
            data.put((byte)mode.getValue());
            data.put(VlqUtils.getVlqBytes(imageSlot1));
            data.put(VlqUtils.getVlqBytes(showTime1));

            final TdpMessage message = new TdpMessage();
            message.setMessageType(ETdpMessageType.COMMAND);
            message.setCommand(ETdpCommand.START_SLIDE_SHOW);
            message.addAddress(1);
            message.setData(data);
            settings = StartSlideShowCommand.parse(message);
            Assert.assertEquals(ETdpSlideShowMode.SHOW_ONCE, settings.getMode());
            Assert.assertEquals(true, settings.getImageSlotsSetting().containsKey(imageSlot1));
            Assert.assertEquals(showTime1, settings.getImageSlotsSetting().get(imageSlot1));

            // correct message
            final ByteBuffer data2 = TdpUtils.createBuffer(9);
            data2.put((byte)mode.getValue());
            data2.put(VlqUtils.getVlqBytes(imageSlot1));
            data2.put(VlqUtils.getVlqBytes(showTime1));
            data2.put(VlqUtils.getVlqBytes(imageSlot2));
            data2.put(VlqUtils.getVlqBytes(showTime2));

            message.setMessageType(ETdpMessageType.COMMAND);
            message.setCommand(ETdpCommand.START_SLIDE_SHOW);
            message.addAddress(1);
            message.setData(data2);
            settings = StartSlideShowCommand.parse(message);
            Assert.assertEquals(ETdpSlideShowMode.SHOW_ONCE, settings.getMode());
            Assert.assertEquals(true, settings.getImageSlotsSetting().containsKey(imageSlot1));
            Assert.assertEquals(showTime1, settings.getImageSlotsSetting().get(imageSlot1));
            Assert.assertEquals(true, settings.getImageSlotsSetting().containsKey(imageSlot2));
            Assert.assertEquals(showTime2, settings.getImageSlotsSetting().get(imageSlot2));
        } catch (TdpProtocolException e) {
            e.printStackTrace();
            Assert.fail("Unexpected TdpProtocolException");
        }
    }

    /**
     * Tests exceptions of the function parse(): Wrong command id
     */
    @Test
    public final void testParseException1() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong command id in command"));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        StartSlideShowCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): invalid message type
     */
    @Test
    public final void testParseException2() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong message type in command"));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.START_SLIDE_SHOW);
        message.addAddress(1);
        StartSlideShowCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): Unknown slide show mode
     */
    @Test
    public final void testParseException3() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Unknown value for slide show mode"));

        // incorrect message
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x04);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.START_SLIDE_SHOW);
        message.addAddress(1);
        message.setData(data);
        StartSlideShowCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): Unknown slide show mode
     */
    @Test
    public final void testParseException4() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No images set for slide show"));

        // incorrect message
        final ByteBuffer data = TdpUtils.createBuffer(1);
        data.put((byte)0x01);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.START_SLIDE_SHOW);
        message.addAddress(1);
        message.setData(data);
        StartSlideShowCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): no data
     */
    @Test
    public final void testParseException5() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("No data in message"));

        // incorrect message (no data)
        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.START_SLIDE_SHOW);
        message.addAddress(1);
        message.setData(null);
        StartSlideShowCommand.parse(message);
    }

    /**
     * Tests exceptions of the function parse(): incomplete message
     */
    @Test
    public final void testParseException7() throws TdpProtocolException {
        thrown.expect(BufferUnderflowException.class);

        final ETdpSlideShowMode mode = ETdpSlideShowMode.SHOW_ONCE;
        final int imageSlot1 = 128;

        // incorrect message
        final ByteBuffer data = TdpUtils.createBuffer(3);
        data.put((byte)mode.getValue());
        data.put(VlqUtils.getVlqBytes(imageSlot1));

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.START_SLIDE_SHOW);
        message.addAddress(1);
        message.setData(data);
        StartSlideShowCommand.parse(message);
    }

    /**
     * Tests the function createData() with maximum values
     */
    @Test
    public final void testCreateDataMaximum() {
        TdpStartSlideShow settings = new TdpStartSlideShow();
        settings.setMode(ETdpSlideShowMode.SHOW_ONCE);
        LinkedHashMap<Integer, Integer> slots = new LinkedHashMap<>();
        slots.put(VlqUtils.MAX_VALUE, VlqUtils.MAX_VALUE);
        settings.setImageSlotsSetting(slots);

        StartSlideShowCommand.createData(settings);
    }
}
