/*
Copyright (c) 2015, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.messages;

import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.List;

import org.disperanto.enums.ETdpCommand;
import org.disperanto.enums.ETdpMessageType;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.utils.TdpUtils;
import org.hamcrest.CoreMatchers;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * JUnit tests for class StartSlideShowResponse.
 */
public class TdpResponseStartSlideShow {

    @Rule
    public ExpectedException thrown = ExpectedException.none();

    /**
     * Tests the function parseData()
     */
    @Test
    public final void testParseData() throws TdpProtocolException {
        final short imageCrc = 127;

        // correct message (only mandatory fields)
        final ByteBuffer data = TdpUtils.createBuffer(2);
        data.putShort(imageCrc); // crc first image

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.START_SLIDE_SHOW);
        message.addAddress(1);
        message.setData(data);

        List<Short> crcList = StartSlideShowResponse.parse(message);

        Assert.assertEquals(imageCrc, (short)crcList.get(0));
    }

    /**
     * Tests the function parseData(): invalid message type
     */
    @Test
    public final void testParseDataException() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong message type"));

        final ByteBuffer data = TdpUtils.createBuffer(0);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.COMMAND);
        message.setCommand(ETdpCommand.START_SLIDE_SHOW);
        message.addAddress(1);
        message.setData(data);

        StartSlideShowResponse.parse(message);
    }

    /**
     * Tests the function parseData(): invalid number of addresses
     */
    @Test
    public final void testParseDataException2() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong number of addresses"));

        final ByteBuffer data = TdpUtils.createBuffer(0);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.START_SLIDE_SHOW);
        message.addAddress(1);
        message.addAddress(2);
        message.setData(data);

        StartSlideShowResponse.parse(message);
    }

    /**
     * Tests the function parseData(): wrong command id
     */
    @Test
    public final void testParseDataException3() throws TdpProtocolException {
        thrown.expect(TdpProtocolException.class);
        thrown.expectMessage(CoreMatchers.containsString("Wrong command id"));

        final ByteBuffer data = TdpUtils.createBuffer(0);

        final TdpMessage message = new TdpMessage();
        message.setMessageType(ETdpMessageType.RESPONSE);
        message.setCommand(ETdpCommand.DISPLAY_PROPERTIES);
        message.addAddress(1);
        message.setData(data);

        StartSlideShowResponse.parse(message);
    }

    /**
     * Tests the function CreateData() with maximum values
     */
    @Test
    public final void testCreateDataMaximum() {
        List<Short> crcList = new ArrayList<Short>();
        crcList.add((short)0xFFFF);
        StartSlideShowResponse.createData(crcList);
    }
}
