/*
Copyright (c) 2013, Technolution B.V.
 All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this 
   list of conditions and the following disclaimer in the documentation and/or 
   other materials provided with the distribution.

3. Neither the name of Technolution nor the names of its contributors may be used 
   to endorse or promote products derived from this software without specific 
   prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE.
*/
package org.disperanto.refapp.prisserver;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.List;

import org.disperanto.communication.UdpSender;
import org.disperanto.data.TdpDisplayProperties;
import org.disperanto.data.TdpNotifications;
import org.disperanto.data.TdpStatus;
import org.disperanto.exception.TdpProtocolException;
import org.disperanto.messages.DisplayPropertiesResponse;
import org.disperanto.messages.StatusResponse;
import org.disperanto.messages.TdpCommandFactory;
import org.disperanto.messages.TdpMessage;
import org.disperanto.protocol.ITdpNotificationListener;
import org.disperanto.protocol.TdpProtocolExchanger;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Handles requests for displays using Disperanto.
 */
public class TdpMobiMaestroService implements ITdpNotificationListener {
    private static final Logger LOGGER = LoggerFactory.getLogger(TdpMobiMaestroService.class);

    // UDP protocol used for sending the Traffic Display messages
    private TdpProtocolExchanger tdpProtocolExchanger;
    private UdpSender connection;

    /**
     * Constructor
     * 
     * @param ipAddress IP address of the sign group of the displays
     * @param port Port number used to send packets
     * @param timeoutMs Timeout for UDP communication in milliseconds
     */
    public TdpMobiMaestroService(String ipAddress, int port, int timeoutMs) throws UnknownHostException {
        LOGGER.info("MobiMaestro service created for IP-address " + ipAddress +
                ", port: " + port + ", timeout: " + timeoutMs);
        // Setup UDP connection
        connection = new UdpSender(ipAddress, port, timeoutMs);

        tdpProtocolExchanger = new TdpProtocolExchanger(connection);
        tdpProtocolExchanger.addNotificationListener(this);

    }

    /**
     * Start the connection
     */
    public final void start() {
        connection.start();
    }

    /**
     * Stops the connection
     */
    public final void stop() {
        connection.stop();
    }

    /**
     * Requests to send a display properties command to one or more displays.
     * 
     * @param displayAddresses List of display addresses the command is sent to
     * @return List of Traffic Display results containing the display properties and notifications
     */
    public final List<TdpDisplayProperties> requestDisplayProperties(List<Integer> displayAddresses)
            throws TdpProtocolException {
        final List<TdpMessage> commands = new ArrayList<TdpMessage>();
        List<TdpMessage> responses = new ArrayList<TdpMessage>();
        final List<TdpDisplayProperties> displayPropertiesList = new ArrayList<TdpDisplayProperties>();

        // Create message for DisplayProperties command
        commands.add(TdpCommandFactory.createDisplayPropertiesCommand(displayAddresses));

        // Communicate with displays
        try {
            responses = tdpProtocolExchanger.exchangeMessages(commands);
        } catch (IOException e) {
            LOGGER.error("IOException: " + e.getMessage(), e);
        }

        if (responses == null) {
            // no response received
            return null;
        }

        // Parse data from responses
        for (TdpMessage response : responses) {
            switch (response.getCommand()) {
            case DISPLAY_PROPERTIES:
                displayPropertiesList.add(DisplayPropertiesResponse.parse(response));
                break;
            default:
                throw new TdpProtocolException("Unexpected command value " + response.getCommand() +
                        " received while requesting display properties");
            }
        }

        return displayPropertiesList;
    }

    /**
     * Requests to send a status command to one or more displays.
     * 
     * @param displayAddresses List of display addresses the command is sent to
     * @return List of Traffic Display results containing the display properties and notifications
     */
    public final List<TdpStatus> requestStatus(List<Integer> displayAddresses) throws TdpProtocolException {
        final List<TdpMessage> commands = new ArrayList<TdpMessage>();
        List<TdpMessage> responses = new ArrayList<TdpMessage>();
        final List<TdpStatus> statusList = new ArrayList<TdpStatus>();

        // Create message for Status command
        commands.add(TdpCommandFactory.createStatusCommand(displayAddresses));

        // Communicate with displays
        try {
            responses = tdpProtocolExchanger.exchangeMessages(commands);
        } catch (IOException e) {
            LOGGER.error("IOException: " + e.getMessage(), e);
        } finally {
            connection.stop();
        }

        // Parse data from responses
        for (TdpMessage response : responses) {
            switch (response.getCommand()) {
            case STATUS:
                statusList.add(StatusResponse.parse(response));
                break;
            default:
                throw new TdpProtocolException("Unexpected command value received while requesting status");
            }
        }

        return statusList;
    }

    @Override
    public final void notificationReceived(TdpNotifications notifications) {
        LOGGER.debug(notifications.toString());
    }
}
